import menpo.landmark.labels as labels
from menpo.landmark import LandmarkGroup
import inspect
import numpy as np
from menpo.shape import PointCloud, bounding_box


EXEMPT_FUNCTIONS = {'labeller', 'bounding_box_to_bounding_box',
                    'bounding_box_mirrored_to_bounding_box'}


def check_label_func(func, input_n_points, output_n_points):
    # Could be any dimensionality
    array = np.zeros([input_n_points, 2])
    pcloud = PointCloud(array)
    lmark_g = LandmarkGroup.init_with_all_label(pcloud)

    array_result = func(array)
    assert isinstance(array_result, PointCloud)
    assert array_result.n_points == output_n_points

    pcloud_result = func(pcloud)
    assert isinstance(pcloud_result, PointCloud)
    assert pcloud_result.n_points == output_n_points

    lmark_g_result = func(lmark_g)
    assert isinstance(lmark_g_result, LandmarkGroup)
    assert lmark_g_result.lms.n_points == output_n_points


def test_labels():
    for fname, func in inspect.getmembers(labels,
                                          predicate=inspect.isfunction):
        if fname not in EXEMPT_FUNCTIONS:
            parts = fname.split('_')
            to_index = parts.index('to')
            input_n_points = parts[to_index - 1]
            output_n_points = parts[-1]
            try:
                input_n_points = int(input_n_points)
            except ValueError:
                # If it isn't the element before 'to', there must be
                # a modifier - so try the element before that
                input_n_points = int(parts[to_index - 2])
            try:
                output_n_points = int(output_n_points)
            except ValueError:
                # If it isn't the last element, there must be
                # a modifier - so try the second to last element
                output_n_points = int(parts[-2])

            yield (check_label_func,
                   func, input_n_points, output_n_points)


def test_bounding_box_to_bounding_box():
    from menpo.landmark import bounding_box_to_bounding_box

    check_label_func(bounding_box_to_bounding_box, 4, 4)


def test_bounding_box_mirrored_to_bounding_box():
    from menpo.landmark import bounding_box_mirrored_to_bounding_box

    check_label_func(bounding_box_mirrored_to_bounding_box, 4, 4)



from collections import OrderedDict
import numpy as np

from ..base import labeller_func, validate_input, connectivity_from_array


@labeller_func(group_label='face_bu3dfe_83')
def face_bu3dfe_83_to_face_bu3dfe_83(pcloud):
    r"""
    Apply the BU-3DFE (Binghamton University 3D Facial Expression)
    Database 83-point facial semantic labels.

    The semantic labels applied are as follows:

      - right_eye
      - left_eye
      - right_eyebrow
      - left_eyebrow
      - right_nose
      - left_nose
      - nostrils
      - outer_mouth
      - inner_mouth
      - jaw

    References
    ----------
    .. [1] http://www.cs.binghamton.edu/~lijun/Research/3DFE/3DFE_Analysis.html
    """
    from menpo.shape import PointUndirectedGraph

    n_expected_points = 83
    validate_input(pcloud, n_expected_points)

    reye_indices = np.arange(0, 8)
    leye_indices = np.arange(8, 16)
    rbrow_indices = np.arange(16, 26)
    lbrow_indices = np.arange(26, 36)
    rnose_indices = np.arange(36, 39)
    lnose_indices = np.arange(39, 42)
    nostril_indices = np.arange(42, 48)
    outermouth_indices = np.arange(48, 60)
    innermouth_indices = np.arange(60, 68)
    jaw_indices = np.arange(68, 83)

    reye_connectivity = connectivity_from_array(reye_indices, close_loop=True)
    leye_connectivity = connectivity_from_array(leye_indices, close_loop=True)
    rbrow_connectivity = connectivity_from_array(rbrow_indices,
                                                 close_loop=True)
    lbrow_connectivity = connectivity_from_array(lbrow_indices,
                                                 close_loop=True)
    rnose_connectivity = connectivity_from_array(rnose_indices)
    nostril_connectivity = connectivity_from_array(nostril_indices)
    lnose_connectivity = connectivity_from_array(lnose_indices)
    outermouth_connectivity = connectivity_from_array(outermouth_indices,
                                                      close_loop=True)
    innermouth_connectivity = connectivity_from_array(innermouth_indices,
                                                      close_loop=True)
    jaw_connectivity = connectivity_from_array(jaw_indices)

    all_connectivity = np.vstack([
        reye_connectivity, leye_connectivity,
        rbrow_connectivity, lbrow_connectivity,
        rnose_connectivity, nostril_connectivity, lnose_connectivity,
        outermouth_connectivity, innermouth_connectivity,
        jaw_connectivity
    ])

    new_pcloud = PointUndirectedGraph.init_from_edges(pcloud.points,
                                                      all_connectivity)

    mapping = OrderedDict()
    mapping['right_eye'] = reye_indices
    mapping['left_eye'] = leye_indices
    mapping['right_eyebrow'] = rbrow_indices
    mapping['left_eyebrow'] = lbrow_indices
    mapping['right_nose'] = rnose_indices
    mapping['left_nose'] = lnose_indices
    mapping['nostrils'] = nostril_indices
    mapping['outer_mouth'] = outermouth_indices
    mapping['inner_mouth'] = innermouth_indices
    mapping['jaw'] = jaw_indices

    return new_pcloud, mapping



from __future__ import division
import random
import warnings

import numpy as np
from numpy.testing import assert_allclose, raises
from nose.plugins.attrib import attr

from menpo.testing import is_same_array
from menpo.image import Image, MaskedImage
from menpo.feature import (hog, lbp, es, igo, daisy, no_op, normalize,
                           normalize_norm, normalize_std, normalize_var)
import menpo.io as mio


def test_imagewindowiterator_hog_padding():
    n_cases = 5
    image_width = np.random.randint(50, 250, [n_cases, 1])
    image_height = np.random.randint(50, 250, [n_cases, 1])
    window_step_horizontal = np.random.randint(1, 10, [n_cases, 1])
    window_step_vertical = np.random.randint(1, 10, [n_cases, 1])
    for i in range(n_cases):
        image = MaskedImage(np.random.randn(1, image_height[i, 0],
                                            image_width[i, 0]))
        hog_im = hog(image, mode='dense',
                     window_step_vertical=window_step_vertical[i, 0],
                     window_step_horizontal=window_step_horizontal[i, 0],
                     window_step_unit='pixels', padding=True)
        n_windows_horizontal = len(range(0, image_width[i, 0],
                                         window_step_horizontal[i, 0]))
        n_windows_vertical = len(range(0, image_height[i, 0],
                                       window_step_vertical[i, 0]))
        assert_allclose(hog_im.shape, (n_windows_vertical,
                                       n_windows_horizontal))


def test_windowiterator_hog_no_padding():
    n_cases = 5
    image_width = np.random.randint(50, 250, [n_cases, 1])
    image_height = np.random.randint(50, 250, [n_cases, 1])
    window_step_horizontal = np.random.randint(1, 10, [n_cases, 1])
    window_step_vertical = np.random.randint(1, 10, [n_cases, 1])
    window_width = np.random.randint(3, 20, [n_cases, 1])
    window_height = np.random.randint(3, 20, [n_cases, 1])
    for i in range(n_cases):
        image = MaskedImage(np.random.randn(1, image_height[i, 0],
                                            image_width[i, 0]))
        hog_img = hog(image, mode='dense', cell_size=3, block_size=1,
                      window_height=window_height[i, 0],
                      window_width=window_width[i, 0], window_unit='pixels',
                      window_step_vertical=window_step_vertical[i, 0],
                      window_step_horizontal=window_step_horizontal[i, 0],
                      window_step_unit='pixels', padding=False)
        n_windows_horizontal = len(range(window_width[i, 0] - 1,
                                         image_width[i, 0],
                                         window_step_horizontal[i, 0]))
        n_windows_vertical = len(range(window_height[i, 0] - 1,
                                       image_height[i, 0],
                                       window_step_vertical[i, 0]))
        assert_allclose(hog_img.shape, (n_windows_vertical,
                                        n_windows_horizontal))


def test_windowiterator_lbp_padding():
    n_cases = 5
    image_width = np.random.randint(50, 250, [n_cases, 1])
    image_height = np.random.randint(50, 250, [n_cases, 1])
    window_step_horizontal = np.random.randint(1, 10, [n_cases, 1])
    window_step_vertical = np.random.randint(1, 10, [n_cases, 1])
    for i in range(n_cases):
        image = MaskedImage(np.random.randn(1, image_height[i, 0],
                                            image_width[i, 0]))
        lbp_img = lbp(image, window_step_vertical=window_step_vertical[i, 0],
                      window_step_horizontal=window_step_horizontal[i, 0],
                      window_step_unit='pixels', padding=True)
        n_windows_horizontal = len(range(0, image_width[i, 0],
                                         window_step_horizontal[i, 0]))
        n_windows_vertical = len(range(0, image_height[i, 0],
                                       window_step_vertical[i, 0]))
        assert_allclose(lbp_img.shape, (n_windows_vertical,
                                        n_windows_horizontal))


def test_windowiterator_lbp_no_padding():
    n_cases = 5
    image_width = np.random.randint(50, 250, [n_cases, 1])
    image_height = np.random.randint(50, 250, [n_cases, 1])
    window_step_horizontal = np.random.randint(1, 10, [n_cases, 1])
    window_step_vertical = np.random.randint(1, 10, [n_cases, 1])
    radius = np.random.randint(3, 5, [n_cases, 1])
    for i in range(n_cases):
        image = Image(np.random.randn(1, image_height[i, 0],
                                      image_width[i, 0]))
        lbp_img = lbp(image, radius=radius[i, 0], samples=8,
                      window_step_vertical=window_step_vertical[i, 0],
                      window_step_horizontal=window_step_horizontal[i, 0],
                      window_step_unit='pixels', padding=False)
        window_size = 2 * radius[i, 0] + 1
        n_windows_horizontal = len(range(window_size - 1, image_width[i, 0],
                                         window_step_horizontal[i, 0]))
        n_windows_vertical = len(range(window_size - 1, image_height[i, 0],
                                       window_step_vertical[i, 0]))
        assert_allclose(lbp_img.shape, (n_windows_vertical,
                                        n_windows_horizontal))


def test_hog_channels_dalaltriggs():
    n_cases = 3
    cell_size = np.random.randint(1, 10, [n_cases, 1])
    block_size = np.random.randint(1, 3, [n_cases, 1])
    num_bins = np.random.randint(7, 9, [n_cases, 1])
    channels = np.random.randint(1, 4, [n_cases, 1])
    for i in range(n_cases):
        image = MaskedImage(np.random.randn(channels[i, 0], 40, 40))
        block_size_pixels = cell_size[i, 0] * block_size[i, 0]
        window_width = np.random.randint(block_size_pixels, 40, 1)
        window_height = np.random.randint(block_size_pixels, 40, 1)
        hog_img = hog(image, mode='dense', algorithm='dalaltriggs',
                      cell_size=cell_size[i, 0], block_size=block_size[i, 0],
                      num_bins=num_bins[i, 0], window_height=window_height[0],
                      window_width=window_width[0], window_unit='pixels',
                      window_step_vertical=3, window_step_horizontal=3,
                      window_step_unit='pixels', padding=True)
        length_per_block = block_size[i, 0] * block_size[i, 0] * num_bins[i, 0]
        n_blocks_horizontal = len(range(block_size_pixels - 1, window_width[0],
                                        cell_size[i, 0]))
        n_blocks_vertical = len(range(block_size_pixels - 1, window_height[0],
                                      cell_size[i, 0]))
        n_channels = n_blocks_horizontal * n_blocks_vertical * length_per_block
        assert_allclose(hog_img.n_channels, n_channels)


def test_hog_channels_zhuramanan():
    n_cases = 3
    cell_size = np.random.randint(2, 10, [n_cases])
    channels = np.random.randint(1, 4, [n_cases])
    for i in range(n_cases):
        image = MaskedImage(np.random.randn(channels[i], 40, 40))
        win_width = np.random.randint(3 * cell_size[i], 40, 1)
        win_height = np.random.randint(3 * cell_size[i], 40, 1)
        hog_img = hog(image, mode='dense', algorithm='zhuramanan',
                      cell_size=cell_size[i],
                      window_height=win_height[0],
                      window_width=win_width[0],
                      window_unit='pixels', window_step_vertical=3,
                      window_step_horizontal=3,
                      window_step_unit='pixels', padding=True, verbose=True)
        length_per_block = 31
        n_blocks_horizontal = np.floor((win_width[0] / cell_size[i]) + 0.5) - 2
        n_blocks_vertical = np.floor((win_height[0] / cell_size[i]) + 0.5) - 2
        n_channels = n_blocks_horizontal * n_blocks_vertical * length_per_block
        assert_allclose(hog_img.n_channels, n_channels)


@attr('cyvlfeat')
def test_dsift_channels():
    from menpo.feature import dsift
    n_cases = 3
    num_bins_horizontal = np.random.randint(1, 3, [n_cases, 1])
    num_bins_vertical = np.random.randint(1, 3, [n_cases, 1])
    num_or_bins = np.random.randint(7, 9, [n_cases, 1])
    cell_size_horizontal = np.random.randint(1, 10, [n_cases, 1])
    cell_size_vertical = np.random.randint(1, 10, [n_cases, 1])
    channels = np.random.randint(1, 4, [n_cases])
    for i in range(n_cases):
        image = MaskedImage(np.random.randn(channels[i], 40, 40))
        dsift_img = dsift(image, window_step_horizontal=1,
                          window_step_vertical=1,
                          num_bins_horizontal=num_bins_horizontal[i, 0],
                          num_bins_vertical=num_bins_vertical[i, 0],
                          num_or_bins=num_or_bins[i, 0],
                          cell_size_horizontal=cell_size_horizontal[i, 0],
                          cell_size_vertical=cell_size_vertical[i, 0])
        n_channels = (num_bins_horizontal[i, 0] * num_bins_vertical[i, 0] *
                      num_or_bins[i, 0])
        assert_allclose(dsift_img.n_channels, n_channels)


def test_lbp_channels():
    n_cases = 3
    n_combs = np.random.randint(1, 6, [n_cases, 1])
    channels = np.random.randint(1, 4, [n_cases, 1])
    for i in range(n_cases):
        radius = random.sample(range(1, 10), n_combs[i, 0])
        samples = random.sample(range(4, 12), n_combs[i, 0])
        image = MaskedImage(np.random.randn(channels[i, 0], 40, 40))
        lbp_img = lbp(image, radius=radius, samples=samples,
                      window_step_vertical=3, window_step_horizontal=3,
                      window_step_unit='pixels', padding=True)
        assert_allclose(lbp_img.n_channels, n_combs[i, 0] * channels[i, 0])


def test_igo_channels():
    n_cases = 3
    channels = np.random.randint(1, 10, [n_cases, 1])
    for i in range(n_cases):
        image = Image(np.random.randn(channels[i, 0], 40, 40))
        igo_img = igo(image)
        igo2_img = igo(image, double_angles=True)
        assert_allclose(igo_img.shape, image.shape)
        assert_allclose(igo2_img.shape, image.shape)
        assert_allclose(igo_img.n_channels, 2 * channels[i, 0])
        assert_allclose(igo2_img.n_channels, 4 * channels[i, 0])


def test_es_channels():
    n_cases = 3
    channels = np.random.randint(1, 10, [n_cases, 1])
    for i in range(n_cases):
        image = Image(np.random.randn(channels[i, 0], 40, 40))
        es_img = es(image)
        assert_allclose(es_img.shape, image.shape)
        assert_allclose(es_img.n_channels, 2 * channels[i, 0])


def test_daisy_channels():
    n_cases = 3
    rings = np.random.randint(1, 3, [n_cases, 1])
    orientations = np.random.randint(1, 7, [n_cases, 1])
    histograms = np.random.randint(1, 6, [n_cases, 1])
    channels = np.random.randint(1, 5, [n_cases, 1])
    for i in range(n_cases):
        image = Image(np.random.randn(channels[i, 0], 40, 40))
        daisy_img = daisy(image, step=4, rings=rings[i, 0],
                          orientations=orientations[i, 0],
                          histograms=histograms[i, 0])
        assert_allclose(daisy_img.shape, (3, 3))
        assert_allclose(daisy_img.n_channels,
                        ((rings[i, 0]*histograms[i, 0]+1)*orientations[i, 0]))


def test_igo_values():
    image = Image([[1., 2.], [2., 1.]])
    igo_img = igo(image)
    res = np.array(
        [[[0.70710678, 0.70710678],
          [-0.70710678, -0.70710678]],
         [[0.70710678, -0.70710678],
          [0.70710678, -0.70710678]]])
    assert_allclose(igo_img.pixels, res)
    image = Image([[0., 0.], [0., 0.]])
    igo_img = igo(image)
    res = np.array([[[0., 0.], [0., 0.]], [[1., 1.], [1., 1.]]])
    assert_allclose(igo_img.pixels, res)


def test_es_values():
    image = Image([[1., 2.], [2., 1.]])
    es_img = es(image)
    k = 1.0 / (2 * (2 ** 0.5))
    res = np.array([[[k, -k], [k, -k]], [[k, k], [-k, -k]]])
    assert_allclose(es_img.pixels, res)


def test_daisy_values():
    image = Image([[1., 2., 3., 4.], [2., 1., 3., 4.], [1., 2., 3., 4.],
                   [2., 1., 3., 4.]])
    daisy_img = daisy(image, step=1, rings=2, radius=1, orientations=8,
                      histograms=8)
    assert_allclose(np.around(daisy_img.pixels[10, 0, 0], 6), 0.001355)
    assert_allclose(np.around(daisy_img.pixels[20, 0, 1], 6), 0.032237)
    assert_allclose(np.around(daisy_img.pixels[30, 1, 0], 6), 0.002032)
    assert_allclose(np.around(daisy_img.pixels[40, 1, 1], 6), 0.000163)


@attr('cyvlfeat')
def test_dsift_values():
    from menpo.feature import dsift
    # Equivalent to the transpose of image in Matlab
    image = Image([[1, 2, 3, 4], [2, 1, 3, 4], [1, 2, 3, 4], [2, 1, 3, 4]])
    sift_img = dsift(image, cell_size_horizontal=2, cell_size_vertical=2)
    assert_allclose(np.around(sift_img.pixels[0, 0, 0], 6), 19.719786,
                    rtol=1e-04)
    assert_allclose(np.around(sift_img.pixels[1, 0, 1], 6), 141.535736,
                    rtol=1e-04)
    assert_allclose(np.around(sift_img.pixels[0, 1, 0], 6), 184.377472,
                    rtol=1e-04)
    assert_allclose(np.around(sift_img.pixels[5, 1, 1], 6), 39.04007,
                    rtol=1e-04)


def test_lbp_values():
    image = Image([[0., 6., 0.], [5., 18., 13.], [0., 20., 0.]])
    lbp_img = lbp(image, radius=1, samples=4, mapping_type='none',
                  padding=False)
    assert_allclose(lbp_img.pixels, 8.)
    image = Image([[0., 6., 0.], [5., 25., 13.], [0., 20., 0.]])
    lbp_img = lbp(image, radius=1, samples=4, mapping_type='riu2',
                  padding=False)
    assert_allclose(lbp_img.pixels, 0.)
    image = Image([[0., 6., 0.], [5., 13., 13.], [0., 20., 0.]])
    lbp_img = lbp(image, radius=1, samples=4, mapping_type='u2', padding=False)
    assert_allclose(lbp_img.pixels, 8.)
    image = Image([[0., 6., 0.], [5., 6., 13.], [0., 20., 0.]])
    lbp_img = lbp(image, radius=1, samples=4, mapping_type='ri', padding=False)
    assert_allclose(lbp_img.pixels, 4.)


def test_constrain_landmarks():
    breaking_bad = mio.import_builtin_asset('breakingbad.jpg').as_masked()
    breaking_bad = breaking_bad.crop_to_landmarks(boundary=20)
    breaking_bad = breaking_bad.resize([50, 50])
    breaking_bad.constrain_mask_to_landmarks()
    hog_b = hog(breaking_bad, mode='sparse')
    x = np.where(hog_b.landmarks['PTS'].lms.points[:, 0] > hog_b.shape[1] - 1)
    y = np.where(hog_b.landmarks['PTS'].lms.points[:, 0] > hog_b.shape[0] - 1)
    assert_allclose(len(x[0]) + len(y[0]), 12)
    hog_b = hog(breaking_bad, mode='sparse')
    hog_b.landmarks['PTS'] = hog_b.landmarks['PTS'].lms.constrain_to_bounds(
        hog_b.bounds())
    x = np.where(hog_b.landmarks['PTS'].lms.points[:, 0] > hog_b.shape[1] - 1)
    y = np.where(hog_b.landmarks['PTS'].lms.points[:, 0] > hog_b.shape[0] - 1)
    assert_allclose(len(x[0]) + len(y[0]), 0)


def test_no_op():
    image = Image([[1., 2.], [2., 1.]])
    new_image = no_op(image)
    assert_allclose(image.pixels, new_image.pixels)
    assert not is_same_array(image.pixels, new_image.pixels)


def test_normalize_no_scale_all():
    pixels = np.arange(27, dtype=np.float).reshape([3, 3, 3])
    image = Image(pixels, copy=False)
    new_image = normalize(image, scale_func=None, mode='all')
    assert_allclose(new_image.pixels, pixels - 13.)


def test_normalize_norm_all():
    pixels = np.arange(27, dtype=np.float).reshape([3, 3, 3])
    image = Image(pixels, copy=False)
    new_image = normalize_norm(image, mode='all')
    assert_allclose(np.linalg.norm(new_image.pixels),  1.)


def test_normalize_norm_channels():
    pixels = np.arange(27, dtype=np.float).reshape([3, 3, 3])
    image = Image(pixels, copy=False)
    new_image = normalize_norm(image, mode='per_channel')
    assert_allclose(np.linalg.norm(new_image.pixels[0]),  1.)
    assert_allclose(np.linalg.norm(new_image.pixels[1]),  1.)
    assert_allclose(np.linalg.norm(new_image.pixels[2]),  1.)


def test_normalize_std_all():
    pixels = np.arange(27, dtype=np.float).reshape([3, 3, 3])
    image = Image(pixels, copy=False)
    new_image = normalize_std(image, mode='all')
    assert_allclose(np.std(new_image.pixels),  1.)


def test_normalize_std_channels():
    pixels = np.arange(27, dtype=np.float).reshape([3, 3, 3])
    image = Image(pixels, copy=False)
    new_image = normalize_std(image, mode='per_channel')
    assert_allclose(np.std(new_image.pixels[0]),  1.)
    assert_allclose(np.std(new_image.pixels[1]),  1.)
    assert_allclose(np.std(new_image.pixels[2]),  1.)


def test_normalize_var_all():
    pixels = np.arange(27, dtype=np.float).reshape([3, 3, 3])
    image = Image(pixels, copy=False)
    new_image = normalize_var(image, mode='all')
    assert_allclose(np.var(new_image.pixels),  0.01648, atol=1e-3)


def test_normalize_var_channels():
    pixels = np.arange(27, dtype=np.float).reshape([3, 3, 3])
    image = Image(pixels, copy=False)
    new_image = normalize_var(image, mode='per_channel')
    assert_allclose(np.var(new_image.pixels[0]),  0.15, atol=1e-5)
    assert_allclose(np.var(new_image.pixels[1]),  0.15, atol=1e-5)
    assert_allclose(np.var(new_image.pixels[2]),  0.15, atol=1e-5)


def test_normalize_no_scale_per_channel():
    pixels = np.arange(27, dtype=np.float).reshape([3, 3, 3])
    image = Image(pixels, copy=False)
    new_image = normalize(image, scale_func=None, mode='per_channel')
    assert_allclose(new_image.pixels[0], pixels[0] - 4.)
    assert_allclose(new_image.pixels[1], pixels[1] - 13.)
    assert_allclose(new_image.pixels[2], pixels[2] - 22.)


def test_normalize_no_scale_per_channel():
    pixels = np.arange(27, dtype=np.float).reshape([3, 3, 3])
    image = Image(pixels, copy=False)
    new_image = normalize(image, scale_func=None, mode='per_channel')
    assert_allclose(new_image.pixels[0], pixels[0] - 4.)
    assert_allclose(new_image.pixels[1], pixels[1] - 13.)
    assert_allclose(new_image.pixels[2], pixels[2] - 22.)


def test_normalize_scale_all():
    pixels = np.arange(27, dtype=np.float).reshape([3, 3, 3])
    dummy_scale = lambda *a, **kwargs: np.array(2.0)
    image = Image(pixels, copy=False)
    new_image = normalize(image, scale_func=dummy_scale, mode='all')
    assert_allclose(new_image.pixels, (pixels - 13.0) / 2.0)


def test_normalize_scale_per_channel():
    pixels = np.arange(27, dtype=np.float).reshape([3, 3, 3])
    image = Image(pixels, copy=False)
    dummy_scale = lambda *a, **kwargs: np.array(2.0)
    new_image = normalize(image, scale_func=dummy_scale, mode='per_channel')
    assert_allclose(new_image.pixels[0], (pixels[0] - 4.) / 2.0)
    assert_allclose(new_image.pixels[1], (pixels[1] - 13.) / 2.0)
    assert_allclose(new_image.pixels[2], (pixels[2] - 22.) / 2.0)


@raises(ValueError)
def test_normalize_unknown_mode_raises():
    image = Image.init_blank((2, 2))
    normalize(image, mode='fake')


@raises(ValueError)
def test_normalize_0_variance_raises():
    image = Image.init_blank((2, 2))
    dummy_scale = lambda *a, **kwargs: np.array(0.0)
    normalize(image, scale_func=dummy_scale)


def test_normalize_0_variance_warning():
    pixels = np.arange(8, dtype=np.float).reshape([2, 2, 2])
    image = Image(pixels, copy=False)
    dummy_scale = lambda *a, **kwargs: np.array([2.0, 0.0])
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        new_image = normalize(image, scale_func=dummy_scale,
                              error_on_divide_by_zero=False,
                              mode='per_channel')
    assert_allclose(new_image.pixels[0], [[-0.75, -0.25], [0.25, 0.75]])
    assert_allclose(new_image.pixels[1], [[-1.5, -0.5], [0.5, 1.5]])





class MultipleAlignment(object):
    r"""
    Abstract base class for aligning multiple `source` shapes to a `target`
    shape.

    Parameters
    ----------
    sources : `list` of :map:`PointCloud`
        List of pointclouds to be aligned.
    target : :map:`PointCloud`, optional
        The target :map:`PointCloud` to align each source to.
        If ``None``, then the mean of the sources is used.

    Raises
    ------
    ValueError
        Need at least two sources to align
    """

    def __init__(self, sources, target=None):
        from menpo.shape import PointCloud
        if len(sources) < 2 and target is None:
            raise ValueError("Need at least two sources to align")
        self.n_sources = len(sources)
        self.n_points, self.n_dims = sources[0].n_points, sources[0].n_dims
        self.sources = sources
        if target is None:
            # set the target to the mean source position
            self.target = PointCloud(
                sum([s.points for s in self.sources]) / self.n_sources)
        else:
            assert self.n_dims, self.n_points == target.shape
            self.target = target



from functools import partial
from warnings import warn
import numpy as np

from menpo.transform import Translation
from .base import Image, _convert_patches_list_to_single_array
from .patches import set_patches


def pwa_point_in_pointcloud(pcloud, indices, batch_size=None):
    """
    Make sure that the decision of whether a point is inside or outside
    the PointCloud is exactly the same as how PWA calculates triangle
    containment. Then, we use the trick of setting the mask to all the
    points that were NOT outside the triangulation. Otherwise, all points
    were inside and we just return those as ``True``. In general, points
    on the boundary are counted as inside the polygon.

    Parameters
    ----------
    pcloud : :map:`PointCloud`
        The pointcloud to use for the containment test.
    indices : (d, n_dims) `ndarray`
        The list of pixel indices to test.
    batch_size : `int` or ``None``, optional
        See constrain_to_pointcloud for more information about the batch_size
        parameter.

    Returns
    -------
    mask : (d,) `bool ndarray`
        Whether each pixel index was in inside the convex hull of the
        pointcloud or not.
    """
    from menpo.transform.piecewiseaffine import PiecewiseAffine
    from menpo.transform.piecewiseaffine import TriangleContainmentError

    try:
        pwa = PiecewiseAffine(pcloud, pcloud)
        pwa.apply(indices, batch_size=batch_size)
        return np.ones(indices.shape[0], dtype=np.bool)
    except TriangleContainmentError as e:
        return ~e.points_outside_source_domain


def convex_hull_point_in_pointcloud(pcloud, indices):
    """
    Uses the matplotlib ``contains_points`` method, which in turn uses:

        "Crossings Multiply algorithm of InsideTest"
        By Eric Haines, 3D/Eye Inc, erich@eye.com
        http://erich.realtimerendering.com/ptinpoly/

    This algorithm uses a per-pixel test and thus tends to produce smoother
    edges. We also guarantee that all points inside PointCloud will be
    included by calculating the **convex hull** of the pointcloud before
    doing the point inside test.

    Points on the boundary are counted as **outside** the polygon.

    Parameters
    ----------
    pcloud : :map:`PointCloud`
        The pointcloud to use for the containment test.
    indices : (d, n_dims) `ndarray`
        The list of pixel indices to test.

    Returns
    -------
    mask : (d,) `bool ndarray`
        Whether each pixel index was in inside the convex hull of the
        pointcloud or not.
    """
    from scipy.spatial import ConvexHull
    from matplotlib.path import Path

    c_hull = ConvexHull(pcloud.points)
    polygon = pcloud.points[c_hull.vertices, :]

    return Path(polygon).contains_points(indices)


class BooleanImage(Image):
    r"""
    A mask image made from binary pixels. The region of the image that is
    left exposed by the mask is referred to as the 'masked region'. The
    set of 'masked' pixels is those pixels corresponding to a ``True`` value in
    the mask.

    Parameters
    ----------
    mask_data : ``(M, N, ..., L)`` `ndarray`
        The binary mask data. Note that there is no channel axis - a 2D Mask
        Image is built from just a 2D numpy array of mask_data.
        Automatically coerced in to boolean values.
    copy: `bool`, optional
        If ``False``, the image_data will not be copied on assignment. Note that
        if the array you provide is not boolean, there **will still be copy**.
        In general this should only be used if you know what you are doing.
    """

    def __init__(self, mask_data, copy=True):
        # Add a channel dimension. We do this little reshape trick to add
        # the axis because this maintains C-contiguous'ness
        mask_data = mask_data.reshape((1,) + mask_data.shape)
        # If we are trying not to copy, but the data we have isn't boolean,
        # then unfortunately, we forced to copy anyway!
        if mask_data.dtype != np.bool:
            mask_data = np.array(mask_data, dtype=np.bool, copy=True,
                                 order='C')
            if not copy:
                warn('The copy flag was NOT honoured. A copy HAS been made. '
                     'Please ensure the data you pass is C-contiguous.')
        super(BooleanImage, self).__init__(mask_data, copy=copy)

    @classmethod
    def init_blank(cls, shape, fill=True, round='ceil', **kwargs):
        r"""
        Returns a blank :map:`BooleanImage` of the requested shape

        Parameters
        ----------
        shape : `tuple` or `list`
            The shape of the image. Any floating point values are rounded
            according to the ``round`` kwarg.
        fill : `bool`, optional
            The mask value to be set everywhere.
        round: ``{ceil, floor, round}``, optional
            Rounding function to be applied to floating point shapes.

        Returns
        -------
        blank_image : :map:`BooleanImage`
            A blank mask of the requested size

        """
        from .base import round_image_shape
        shape = round_image_shape(shape, round)
        if fill:
            mask = np.ones(shape, dtype=np.bool)
        else:
            mask = np.zeros(shape, dtype=np.bool)
        return cls(mask, copy=False)

    @classmethod
    def init_from_pointcloud(cls, pointcloud, group=None, boundary=0,
                             constrain=True, fill=True):
        r"""
        Create an Image that is big enough to contain the given pointcloud.
        The pointcloud will be translated to the origin and then translated
        according to its bounds in order to fit inside the new image.
        An optional boundary can be provided in order to increase the space
        around the boundary of the pointcloud. The boundary will be added
        to *all sides of the image* and so a boundary of 5 provides 10 pixels
        of boundary total for each dimension.

        By default, the mask will be constrained to the convex hull of the
        provided pointcloud.

        Parameters
        ----------
        pointcloud : :map:`PointCloud`
            Pointcloud to place inside the newly created image.
        group : `str`, optional
            If ``None``, the pointcloud will only be used to create the image.
            If a `str` then the pointcloud will be attached as a landmark
            group to the image, with the given string as key.
        boundary : `float`
            A optional padding distance that is added to the pointcloud bounds.
            Default is ``0``, meaning the max/min of tightest possible
            containing image is returned.
        fill : `int`, optional
            The value to fill all pixels with.
        constrain : `bool`, optional
            If ``True``, the ``True`` values will be image will be constrained
            to the convex hull of the provided pointcloud. If ``False``,
            the mask will be the value of ``fill``.

        Returns
        -------
        image : :map:`MaskedImage`
            A new image with the same size as the given pointcloud, optionally
            with the pointcloud attached as landmarks and the mask constrained
            to the convex hull of the pointcloud.
        """
        # Translate pointcloud to the origin
        minimum = pointcloud.bounds(boundary=boundary)[0]
        origin_pc = Translation(-minimum).apply(pointcloud)
        image_shape = origin_pc.range(boundary=boundary)
        new_image = cls.init_blank(image_shape, fill=fill)
        if constrain:
            new_image = new_image.constrain_to_pointcloud(origin_pc)
        if group is not None:
            new_image.landmarks[group] = origin_pc
        return new_image

    def as_masked(self, mask=None, copy=True):
        r"""
        Impossible for a :map:`BooleanImage` to be transformed to a
        :map:`MaskedImage`.
        """
        raise NotImplementedError("as_masked cannot be invoked on a "
                                  "BooleanImage.")

    @property
    def mask(self):
        r"""
        Returns the pixels of the mask with no channel axis. This is what
        should be used to mask any k-dimensional image.

        :type: ``(M, N, ..., L)``, `bool ndarray`
        """
        return self.pixels[0, ...]

    def n_true(self):
        r"""
        The number of ``True`` values in the mask.

        :type: `int`
        """
        return np.sum(self.pixels)

    def n_false(self):
        r"""
        The number of ``False`` values in the mask.

        :type: `int`
        """
        return self.n_pixels - self.n_true()

    def all_true(self):
        r"""
        ``True`` iff every element of the mask is ``True``.

        :type: `bool`
        """
        return np.all(self.pixels)

    def proportion_true(self):
        r"""
        The proportion of the mask which is ``True``.

        :type: `float`
        """
        return (self.n_true() * 1.0) / self.n_pixels

    def proportion_false(self):
        r"""
        The proportion of the mask which is ``False``

        :type: `float`
        """
        return (self.n_false() * 1.0) / self.n_pixels

    def true_indices(self):
        r"""
        The indices of pixels that are ``True``.

        :type: ``(n_dims, n_true)`` `ndarray`
        """
        if self.all_true():
            return self.indices()
        else:
            # Ignore the channel axis
            return np.vstack(np.nonzero(self.pixels[0])).T

    def false_indices(self):
        r"""
        The indices of pixels that are ``Flase``.

        :type: ``(n_dims, n_false)`` `ndarray`
        """
        # Ignore the channel axis
        return np.vstack(np.nonzero(~self.pixels[0])).T

    def __str__(self):
        return ('{} {}D mask, {:.1%} '
                'of which is True'.format(self._str_shape(), self.n_dims,
                                          self.proportion_true()))

    def from_vector(self, vector, copy=True):
        r"""
        Takes a flattened vector and returns a new :map:`BooleanImage` formed
        by reshaping the vector to the correct dimensions. Note that this is
        rebuilding a boolean image **itself** from boolean values. The mask
        is in no way interpreted in performing the operation, in contrast to
        :map:`MaskedImage`, where only the masked region is used in
        :meth:`from_vector` and :meth`as_vector`. Any image landmarks are
        transferred in the process.

        Parameters
        ----------
        vector : ``(n_pixels,)`` `bool ndarray`
            A flattened vector of all the pixels of a :map:`BooleanImage`.
        copy : `bool`, optional
            If ``False``, no copy of the vector will be taken.

        Returns
        -------
        image : :map:`BooleanImage`
            New BooleanImage of same shape as this image

        Raises
        ------
        Warning
            If ``copy=False`` cannot be honored.
        """
        mask = BooleanImage(vector.reshape(self.shape), copy=copy)
        if self.has_landmarks:
            mask.landmarks = self.landmarks
        if hasattr(self, 'path'):
            mask.path = self.path
        return mask

    def invert(self):
        r"""
        Returns a copy of this boolean image, which is inverted.

        Returns
        -------
        inverted : :map:`BooleanImage`
            A copy of this boolean mask, where all ``True`` values are ``False``
            and all ``False`` values are ``True``.
        """
        inverse = self.copy()
        inverse.pixels = ~self.pixels
        return inverse

    def bounds_true(self, boundary=0, constrain_to_bounds=True):
        r"""
        Returns the minimum to maximum indices along all dimensions that the
        mask includes which fully surround the ``True`` mask values. In the case
        of a 2D Image for instance, the min and max define two corners of a
        rectangle bounding the True pixel values.

        Parameters
        ----------
        boundary : `int`, optional
            A number of pixels that should be added to the extent. A
            negative value can be used to shrink the bounds in.
        constrain_to_bounds: `bool`, optional
            If ``True``, the bounding extent is snapped to not go beyond
            the edge of the image. If ``False``, the bounds are left unchanged.

        Returns
        --------
        min_b : ``(D,)`` `ndarray`
            The minimum extent of the ``True`` mask region with the boundary
            along each dimension. If ``constrain_to_bounds=True``,
            is clipped to legal image bounds.
        max_b : ``(D,)`` `ndarray`
            The maximum extent of the ``True`` mask region with the boundary
            along each dimension. If ``constrain_to_bounds=True``,
            is clipped to legal image bounds.
        """
        mpi = self.true_indices()
        maxes = np.max(mpi, axis=0) + boundary
        mins = np.min(mpi, axis=0) - boundary
        if constrain_to_bounds:
            maxes = self.constrain_points_to_bounds(maxes)
            mins = self.constrain_points_to_bounds(mins)
        return mins, maxes

    def bounds_false(self, boundary=0, constrain_to_bounds=True):
        r"""
        Returns the minimum to maximum indices along all dimensions that the
        mask includes which fully surround the False mask values. In the case
        of a 2D Image for instance, the min and max define two corners of a
        rectangle bounding the False pixel values.

        Parameters
        ----------
        boundary : `int` >= 0, optional
            A number of pixels that should be added to the extent. A
            negative value can be used to shrink the bounds in.
        constrain_to_bounds: `bool`, optional
            If ``True``, the bounding extent is snapped to not go beyond
            the edge of the image. If ``False``, the bounds are left unchanged.

        Returns
        -------
        min_b : ``(D,)`` `ndarray`
            The minimum extent of the ``True`` mask region with the boundary
            along each dimension. If ``constrain_to_bounds=True``,
            is clipped to legal image bounds.
        max_b : ``(D,)`` `ndarray`
            The maximum extent of the ``True`` mask region with the boundary
            along each dimension. If ``constrain_to_bounds=True``,
            is clipped to legal image bounds.
        """
        return self.invert().bounds_true(
            boundary=boundary, constrain_to_bounds=constrain_to_bounds)

    # noinspection PyMethodOverriding
    def sample(self, points_to_sample, mode='constant', cval=False, **kwargs):
        r"""
        Sample this image at the given sub-pixel accurate points. The input
        PointCloud should have the same number of dimensions as the image e.g.
        a 2D PointCloud for a 2D multi-channel image. A numpy array will be
        returned the has the values for every given point across each channel
        of the image.

        Parameters
        ----------
        points_to_sample : :map:`PointCloud`
            Array of points to sample from the image. Should be
            `(n_points, n_dims)`
        mode : ``{constant, nearest, reflect, wrap}``, optional
            Points outside the boundaries of the input are filled according
            to the given mode.
        cval : `float`, optional
            Used in conjunction with mode ``constant``, the value outside
            the image boundaries.

        Returns
        -------
        sampled_pixels : (`n_points`, `n_channels`) `bool ndarray`
            The interpolated values taken across every channel of the image.
        """
        # enforce the order as 0, as this is boolean data, then call super
        return Image.sample(self, points_to_sample, order=0, mode=mode,
                            cval=cval)

    # noinspection PyMethodOverriding
    def warp_to_mask(self, template_mask, transform, warp_landmarks=True,
                     mode='constant', cval=False, batch_size=None,
                     return_transform=False):
        r"""
        Return a copy of this :map:`BooleanImage` warped into a different
        reference space.

        Note that warping into a mask is slower than warping into a full image.
        If you don't need a non-linear mask, consider warp_to_shape instead.

        Parameters
        ----------
        template_mask : :map:`BooleanImage`
            Defines the shape of the result, and what pixels should be
            sampled.
        transform : :map:`Transform`
            Transform **from the template space back to this image**.
            Defines, for each pixel location on the template, which pixel
            location should be sampled from on this image.
        warp_landmarks : `bool`, optional
            If ``True``, result will have the same landmark dictionary
            as self, but with each landmark updated to the warped position.
        mode : ``{constant, nearest, reflect or wrap}``, optional
            Points outside the boundaries of the input are filled according
            to the given mode.
        cval : `float`, optional
            Used in conjunction with mode ``constant``, the value outside
            the image boundaries.
        batch_size : `int` or ``None``, optional
            This should only be considered for large images. Setting this
            value can cause warping to become much slower, particular for
            cached warps such as Piecewise Affine. This size indicates
            how many points in the image should be warped at a time, which
            keeps memory usage low. If ``None``, no batching is used and all
            points are warped at once.
        return_transform : `bool`, optional
            This argument is for internal use only. If ``True``, then the
            :map:`Transform` object is also returned.

        Returns
        -------
        warped_image : :map:`BooleanImage`
            A copy of this image, warped.
        transform : :map:`Transform`
            The transform that was used. It only applies if
            `return_transform` is ``True``.
        """
        # enforce the order as 0, as this is boolean data, then call super
        return Image.warp_to_mask(
            self, template_mask, transform, warp_landmarks=warp_landmarks,
            order=0, mode=mode, cval=cval, batch_size=batch_size,
            return_transform=return_transform)

    # noinspection PyMethodOverriding
    def warp_to_shape(self, template_shape, transform, warp_landmarks=True,
                      mode='constant', cval=False, order=None,
                      batch_size=None, return_transform=False):
        """
        Return a copy of this :map:`BooleanImage` warped into a different
        reference space.

        Note that the order keyword argument is in fact ignored, as any order
        other than 0 makes no sense on a binary image. The keyword argument is
        present only for compatibility with the :map:`Image` warp_to_shape API.

        Parameters
        ----------
        template_shape : ``(n_dims, )`` `tuple` or `ndarray`
            Defines the shape of the result, and what pixel indices should be
            sampled (all of them).
        transform : :map:`Transform`
            Transform **from the template_shape space back to this image**.
            Defines, for each index on template_shape, which pixel location
            should be sampled from on this image.
        warp_landmarks : `bool`, optional
            If ``True``, result will have the same landmark dictionary
            as self, but with each landmark updated to the warped position.
        mode : ``{constant, nearest, reflect or wrap}``, optional
            Points outside the boundaries of the input are filled according
            to the given mode.
        cval : `float`, optional
            Used in conjunction with mode ``constant``, the value outside
            the image boundaries.
        batch_size : `int` or ``None``, optional
            This should only be considered for large images. Setting this
            value can cause warping to become much slower, particular for
            cached warps such as Piecewise Affine. This size indicates
            how many points in the image should be warped at a time, which
            keeps memory usage low. If ``None``, no batching is used and all
            points are warped at once.
        return_transform : `bool`, optional
            This argument is for internal use only. If ``True``, then the
            :map:`Transform` object is also returned.

        Returns
        -------
        warped_image : :map:`BooleanImage`
            A copy of this image, warped.
        transform : :map:`Transform`
            The transform that was used. It only applies if
            `return_transform` is ``True``.
        """
        # call the super variant and get ourselves an Image back
        # note that we force the use of order=0 for BooleanImages.
        warped = Image.warp_to_shape(self, template_shape, transform,
                                     warp_landmarks=warp_landmarks, order=0,
                                     mode=mode, cval=cval,
                                     batch_size=batch_size)
        # unfortunately we can't escape copying here, let BooleanImage
        # convert us to np.bool
        boolean_image = BooleanImage(warped.pixels.reshape(template_shape))
        if warped.has_landmarks:
            boolean_image.landmarks = warped.landmarks
        if hasattr(warped, 'path'):
            boolean_image.path = warped.path
        # optionally return the transform
        if return_transform:
            return boolean_image, transform
        else:
            return boolean_image

    def _build_warp_to_mask(self, template_mask, sampled_pixel_values,
                            **kwargs):
        r"""
        Builds the warped image from the template mask and sampled pixel values.
        """
        # start from a copy of the template_mask
        warped_img = template_mask.copy()
        if warped_img.all_true():
            # great, just reshape the sampled_pixel_values
            warped_img.pixels = sampled_pixel_values.reshape(
                (1,) + warped_img.shape)
        else:
            # we have to fill out mask with the sampled mask..
            warped_img.pixels[:, warped_img.mask] = sampled_pixel_values
        return warped_img

    def constrain_to_landmarks(self, group=None, batch_size=None):
        r"""
        Returns a copy of this image whereby the ``True`` values in the image
        are restricted to be equal to the convex hull around the landmarks
        chosen. This is not a per-pixel convex hull, but instead relies on a
        triangulated approximation. If the landmarks in question are an instance
        of :map:`TriMesh`, the triangulation of the landmarks will be used in
        the convex hull calculation. If the landmarks are an instance of
        :map:`PointCloud`, Delaunay triangulation will be used to create a
        triangulation.

        Parameters
        ----------
        group : `str`, optional
            The key of the landmark set that should be used. If ``None``,
            and if there is only one set of landmarks, this set will be used.
        batch_size : `int` or ``None``, optional
            This should only be considered for large images. Setting this value
            will cause constraining to become much slower. This size indicates
            how many points in the image should be checked at a time, which
            keeps memory usage low. If ``None``, no batching is used and all
            points are checked at once.

        Returns
        -------
        constrained : :map:`BooleanImage`
            The new boolean image, constrained by the given landmark group.
        """
        return self.constrain_to_pointcloud(self.landmarks[group].lms,
                                            batch_size=batch_size)

    def constrain_to_pointcloud(self, pointcloud, batch_size=None,
                                point_in_pointcloud='pwa'):
        r"""
        Returns a copy of this image whereby the ``True`` values in the image
        are restricted to be equal to the convex hull around a pointcloud. The
        choice of whether a pixel is inside or outside of the pointcloud is
        determined by the ``point_in_pointcloud`` parameter. By default a
        Piecewise Affine transform is used to test for containment, which is
        useful when aligning images by their landmarks. Triangluation will be
        decided by Delauny - if you wish to customise it, a :map:`TriMesh`
        instance can be passed for the ``pointcloud`` argument. In this case,
        the triangulation of the Trimesh will be used to define the retained
        region.

        For large images, a faster and pixel-accurate method can be used (
        'convex_hull'). Here, there is no specialization for
        :map:`TriMesh` instances. Alternatively, a callable can be provided to
        override the test. By default, the provided implementations are only
        valid for 2D images.


        Parameters
        ----------
        pointcloud : :map:`PointCloud` or :map:`TriMesh`
            The pointcloud of points that should be constrained to. See
            `point_in_pointcloud` for how in some cases a :map:`TriMesh` may be
            used to control triangulation.
        batch_size : `int` or ``None``, optional
            This should only be considered for large images. Setting this value
            will cause constraining to become much slower. This size indicates
            how many points in the image should be checked at a time, which
            keeps memory usage low. If ``None``, no batching is used and all
            points are checked at once. By default, this is only used for
            the 'pwa' point_in_pointcloud choice.
        point_in_pointcloud : {'pwa', 'convex_hull'} or `callable`
            The method used to check if pixels in the image fall inside the
            ``pointcloud`` or not. If 'pwa', Menpo's :map:`PiecewiseAffine`
            transform will be used to test for containment. In this case
            ``pointcloud`` should be a :map:`TriMesh`. If it isn't, Delauny
            triangulation will be used to first triangulate ``pointcloud`` into
            a  :map:`TriMesh` before testing for containment.
            If a callable is passed, it should take two parameters,
            the :map:`PointCloud` to constrain with and the pixel locations
            ((d, n_dims) ndarray) to test and should return a (d, 1) boolean
            ndarray of whether the pixels were inside (True) or outside (False)
            of the :map:`PointCloud`.

        Returns
        -------
        constrained : :map:`BooleanImage`
            The new boolean image, constrained by the given pointcloud.

        Raises
        ------
        ValueError
            If the image is not 2D and a default implementation is chosen.
        ValueError
            If the chosen ``point_in_pointcloud`` is unknown.
        """
        copy = self.copy()
        if point_in_pointcloud in {'pwa', 'convex_hull'} and self.n_dims != 2:
            raise ValueError('Can only constrain mask on 2D images with the '
                             'default point_in_pointcloud implementations.'
                             'Please provide a custom callable for calculating '
                             'the new mask in this '
                             '{}D image'.format(self.n_dims))

        if point_in_pointcloud == 'pwa':
            point_in_pointcloud = partial(pwa_point_in_pointcloud,
                                          batch_size=batch_size)
        elif point_in_pointcloud == 'convex_hull':
            point_in_pointcloud = convex_hull_point_in_pointcloud
        elif not callable(point_in_pointcloud):
            # Not a function, or a string, so we have an error!
            raise ValueError('point_in_pointcloud must be a callable that '
                             'take two arguments: the Menpo PointCloud as a '
                             'boundary and the ndarray of pixel indices '
                             'to test. {} is an unknown option.'.format(
                             point_in_pointcloud))

        # Only consider indices inside the bounding box of the PointCloud
        bounds = pointcloud.bounds()
        # Convert to integer to try and reduce boundary fp rounding errors.
        bounds = [b.astype(np.int) for b in bounds]
        indices = copy.indices()

        # This loop is to ensure the code is multi-dimensional
        for k in range(self.n_dims):
            indices = indices[indices[:, k] >= bounds[0][k], :]
            indices = indices[indices[:, k] <= bounds[1][k], :]
        # Due to only testing bounding box indices, make sure the mask starts
        # off as all False
        copy.pixels[:] = False

        # slice(0, 1) because we know we only have 1 channel
        # Slice all the channels, only inside the bounding box (for setting
        # the new mask values).
        all_channels = [slice(0, 1)]
        slices = all_channels + [slice(bounds[0][k], bounds[1][k] + 1)
                                 for k in range(self.n_dims)]
        copy.pixels[slices].flat = point_in_pointcloud(pointcloud, indices)
        return copy

    def set_patches(self, patches, patch_centers, offset=None,
                    offset_index=None):
        r"""
        Set the values of a group of patches into the correct regions in a copy
        of this image. Given an array of patches and a set of patch centers,
        the patches' values are copied in the regions of the image that are
        centred on the coordinates of the given centers.

        The patches argument can have any of the two formats that are returned
        from the `extract_patches()` and `extract_patches_around_landmarks()`
        methods. Specifically it can be:

            1. ``(n_center, n_offset, self.n_channels, patch_shape)`` `ndarray`
            2. `list` of ``n_center * n_offset`` :map:`Image` objects

        Currently only 2D images are supported.

        Parameters
        ----------
        patches : `ndarray` or `list`
            The values of the patches. It can have any of the two formats that
            are returned from the `extract_patches()` and
            `extract_patches_around_landmarks()` methods. Specifically, it can
            either be an ``(n_center, n_offset, self.n_channels, patch_shape)``
            `ndarray` or a `list` of ``n_center * n_offset`` :map:`Image`
            objects.
        patch_centers : :map:`PointCloud`
            The centers to set the patches around.
        offset : `list` or `tuple` or ``(1, 2)`` `ndarray` or ``None``, optional
            The offset to apply on the patch centers within the image.
            If ``None``, then ``(0, 0)`` is used.
        offset_index : `int` or ``None``, optional
            The offset index within the provided `patches` argument, thus the
            index of the second dimension from which to sample. If ``None``,
            then ``0`` is used.

        Raises
        ------
        ValueError
            If image is not 2D
        ValueError
            If offset does not have shape (1, 2)

        Returns
        -------
        new_image : :map:`BooleanImage`
            A new boolean image where the provided patch locations have been
            set to the provided values.
        """
        # parse arguments
        if self.n_dims != 2:
            raise ValueError('Only two dimensional patch insertion is '
                             'currently supported.')
        if offset is None:
            offset = np.zeros([1, 2], dtype=np.intp)
        elif isinstance(offset, tuple) or isinstance(offset, list):
            offset = np.asarray([offset])
        offset = np.require(offset, dtype=np.intp)
        if not offset.shape == (1, 2):
            raise ValueError('The offset must be a tuple, a list or a '
                             'numpy.array with shape (1, 2).')
        if offset_index is None:
            offset_index = 0

        # if patches is a list, convert it to array
        if isinstance(patches, list):
            patches = _convert_patches_list_to_single_array(
                patches, patch_centers.n_points)

        copy = self.copy()
        # convert pixels to uint8 so that they get recognized by cython
        tmp_pixels = copy.pixels.astype(np.uint8)
        # convert patches to uint8 as well and set them to pixels
        set_patches(patches.astype(np.uint8), tmp_pixels, patch_centers.points,
                    offset, offset_index)
        # convert pixels back to bool
        copy.pixels = tmp_pixels.astype(np.bool)
        return copy



from __future__ import division
import numpy as np

from menpo.base import doc_inherit, name_of_callable
from menpo.math import pca, pcacov, ipca, as_matrix
from .linear import MeanLinearVectorModel
from .vectorizable import VectorizableBackedModel


class PCAVectorModel(MeanLinearVectorModel):
    r"""
    A :map:`MeanLinearModel` where components are Principal Components.

    Principal Component Analysis (PCA) by eigenvalue decomposition of the
    data's scatter matrix. For details of the implementation of PCA, see
    :map:`pca`.

    Parameters
    ----------
    samples : `ndarray` or `list` or `iterable` of `ndarray`
        List or iterable of numpy arrays to build the model from, or an
        existing data matrix.
    centre : `bool`, optional
        When ``True`` (default) PCA is performed after mean centering the data.
        If ``False`` the data is assumed to be centred, and the mean will be
        ``0``.
    n_samples : `int`, optional
        If provided then ``samples`` must be an iterator that yields
        ``n_samples``. If not provided then samples has to be a `list` (so we
        know how large the data matrix needs to be).
    max_n_components : `int`, optional
        The maximum number of components to keep in the model. Any components
        above and beyond this one are discarded.
    inplace : `bool`, optional
        If ``True`` the data matrix is modified in place. Otherwise, the data
        matrix is copied.
    """
    def __init__(self, samples, centre=True, n_samples=None,
                 max_n_components=None, inplace=True):
        # Generate data matrix
        data, self.n_samples = self._data_to_matrix(samples, n_samples)

        # Compute pca
        e_vectors, e_values, mean = pca(data, centre=centre, inplace=inplace)

        # The call to __init__ of MeanLinearModel is done in here
        self._constructor_helper(
            eigenvalues=e_values, eigenvectors=e_vectors, mean=mean,
            centred=centre, max_n_components=max_n_components)

    @classmethod
    def init_from_covariance_matrix(cls, C, mean, n_samples, centred=True,
                                    is_inverse=False, max_n_components=None):
        r"""
        Build the Principal Component Analysis (PCA) by eigenvalue
        decomposition of the provided covariance/scatter matrix. For details
        of the implementation of PCA, see :map:`pcacov`.

        Parameters
        ----------
        C : ``(n_features, n_features)`` `ndarray` or `scipy.sparse`
            The Covariance/Scatter matrix. If it is a precision matrix (inverse
            covariance), then set `is_inverse=True`.
        mean : ``(n_features, )`` `ndarray`
            The mean vector.
        n_samples : `int`
            The number of samples used to generate the covariance matrix.
        centred : `bool`, optional
            When ``True`` we assume that the data were centered before
            computing the covariance matrix.
        is_inverse : `bool`, optional
            It ``True``, then it is assumed that `C` is a precision matrix (
            inverse covariance). Thus, the eigenvalues will be inverted. If
            ``False``, then it is assumed that `C` is a covariance matrix.
        max_n_components : `int`, optional
            The maximum number of components to keep in the model. Any
            components above and beyond this one are discarded.
        """
        # Compute pca on covariance
        e_vectors, e_values = pcacov(C, is_inverse=is_inverse)

        # Create new pca instance
        model = PCAModel.__new__(cls)
        model.n_samples = n_samples

        # The call to __init__ of MeanLinearModel is done in here
        model._constructor_helper(
            eigenvalues=e_values, eigenvectors=e_vectors, mean=mean,
            centred=centred, max_n_components=max_n_components)
        return model

    @classmethod
    def init_from_components(cls, components, eigenvalues, mean, n_samples,
                             centred, max_n_components=None):
        r"""
        Build the Principal Component Analysis (PCA) using the provided
        components (eigenvectors) and eigenvalues.

        Parameters
        ----------
        components : ``(n_components, n_features)`` `ndarray`
            The eigenvectors to be used.
        eigenvalues : ``(n_components, )`` `ndarray`
            The corresponding eigenvalues.
        mean : ``(n_features, )`` `ndarray`
            The mean vector.
        n_samples : `int`
            The number of samples used to generate the eigenvectors.
        centred : `bool`, optional
            When ``True`` we assume that the data were centered before
            computing the eigenvectors.
        max_n_components : `int`, optional
            The maximum number of components to keep in the model. Any
            components above and beyond this one are discarded.
        """
        # This is a bit of a filthy trick that by rights should not be done,
        # but we want to have these nice static constructors so we are living
        # with the shame (create an empty object instance which we fill in).
        model = PCAModel.__new__(cls)
        model.n_samples = n_samples

        # The call to __init__ of MeanLinearModel is done in here
        model._constructor_helper(
            eigenvalues=eigenvalues, eigenvectors=components, mean=mean,
            centred=centred, max_n_components=max_n_components)
        return model

    def _constructor_helper(self, eigenvalues, eigenvectors, mean, centred,
                            max_n_components):
        # if covariance is not centred, mean must be zeros.
        if centred:
            MeanLinearVectorModel.__init__(self, eigenvectors, mean)
        else:
            MeanLinearVectorModel.__init__(self, eigenvectors,
                                           np.zeros(mean.shape, dtype=mean.dtype))
        self.centred = centred
        self._eigenvalues = eigenvalues
        # start the active components as all the components
        self._n_active_components = int(self.n_components)
        self._trimmed_eigenvalues = np.array([])
        if max_n_components is not None:
            self.trim_components(max_n_components)

    def _data_to_matrix(self, data, n_samples):
        # build a data matrix from all the samples
        if n_samples is None:
            n_samples = len(data)
        # Assumed data is ndarray of (n_samples, n_features) or list of samples
        if not isinstance(data, np.ndarray):
            # Make sure we have an array, slice of the number of requested
            # samples
            data = np.array(data)[:n_samples]
        return data, n_samples

    @property
    def n_active_components(self):
        r"""
        The number of components currently in use on this model.

        :type: `int`
        """
        return self._n_active_components

    @n_active_components.setter
    def n_active_components(self, value):
        r"""
        Sets an updated number of active components on this model. The number
        of active components represents the number of principal components
        that will be used for generative purposes. Note that this therefore
        makes the model stateful. Also note that setting the number of
        components will not affect memory unless :meth:`trim_components`
        is called.

        Parameters
        ----------
        value : `int`
            The new number of active components.

        Raises
        ------
        ValueError
            Tried setting n_active_components to {value} - value needs to be a
            float 0.0 < n_components < self._total_kept_variance_ratio ({}) or
            an integer 1 < n_components < self.n_components ({})
        """
        err_str = ("Tried setting n_active_components to {} - "
                   "value needs to be a float "
                   "0.0 < n_components < self._total_kept_variance_ratio "
                   "({}) or an integer 1 < n_components < "
                   "self.n_components ({})".format(
            value, self._total_variance_ratio(), self.n_components))

        # check value
        if isinstance(value, float):
            if 0.0 < value <= self._total_variance_ratio():
                # value needed to capture desired variance
                value = np.sum(
                    [r < value
                     for r in self._total_eigenvalues_cumulative_ratio()]) + 1
            else:
                # variance must be bigger than 0.0
                raise ValueError(err_str)
        if isinstance(value, int):
            if value < 1:
                # at least 1 value must be kept
                raise ValueError(err_str)
            elif value >= self.n_components:
                if self.n_active_components < self.n_components:
                    # if the number of available components is smaller than
                    # the total number of components set value to the later
                    value = self.n_components
                else:
                    # if the previous is false and value bigger than the
                    # total number of components, do nothing
                    return
        if 0 < value <= self.n_components:
            self._n_active_components = int(value)
        else:
            raise ValueError(err_str)

    @MeanLinearVectorModel.components.getter
    def components(self):
        r"""
        Returns the active components of the model.

        :type: ``(n_active_components, n_features)`` `ndarray`
        """
        return self._components[:self.n_active_components, :]

    @property
    def eigenvalues(self):
        r"""
        Returns the eigenvalues associated with the active components of the
        model, i.e. the amount of variance captured by each active component,
        sorted form largest to smallest.

        :type: ``(n_active_components,)`` `ndarray`
        """
        return self._eigenvalues[:self.n_active_components]

    def whitened_components(self):
        r"""
        Returns the active components of the model, whitened.

        Returns
        -------
        whitened_components : ``(n_active_components, n_features)`` `ndarray`
            The whitened components.
        """
        return self.components / (
            np.sqrt(self.eigenvalues * self.n_samples +
                    self.noise_variance())[:, None])

    def original_variance(self):
        r"""
        Returns the total amount of variance captured by the original model,
        i.e. the amount of variance present on the original samples.

        Returns
        -------
        optional_variance : `float`
            The variance captured by the model.
        """
        return self._eigenvalues.sum() + self._trimmed_eigenvalues.sum()

    def variance(self):
        r"""
        Returns the total amount of variance retained by the active
        components.

        Returns
        -------
        variance : `float`
            Total variance captured by the active components.
        """
        return self.eigenvalues.sum()

    def _total_variance(self):
        r"""
        Returns the total amount of variance retained by all components
        (active and inactive). Useful when the model has been trimmed.

        Returns
        -------
        total_variance : `float`
            Total variance captured by all components.
        """
        return self._eigenvalues.sum()

    def variance_ratio(self):
        r"""
        Returns the ratio between the amount of variance retained by the
        active components and the total amount of variance present on the
        original samples.

        Returns
        -------
        variance_ratio : `float`
            Ratio of active components variance and total variance present
            in original samples.
        """
        return self.variance() / self.original_variance()

    def _total_variance_ratio(self):
        r"""
        Returns the ratio between the total amount of variance retained by
        all components (active and inactive) and the total amount of variance
        present on the original samples.

        Returns
        -------
        total_variance_ratio : `float`
            Ratio of total variance over the original variance.
        """
        return self._total_variance() / self.original_variance()

    def eigenvalues_ratio(self):
        r"""
        Returns the ratio between the variance captured by each active
        component and the total amount of variance present on the original
        samples.

        Returns
        -------
        eigenvalues_ratio : ``(n_active_components,)`` `ndarray`
            The active eigenvalues array scaled by the original variance.
        """
        return self.eigenvalues / self.original_variance()

    def _total_eigenvalues_ratio(self):
        r"""
        Returns the ratio between the variance captured by each active
        component and the total amount of variance present on the original
        samples.

        Returns
        -------
        total_eigenvalues_ratio : ``(n_components,)`` `ndarray`
            Array of eigenvalues scaled by the original variance.
        """
        return self._eigenvalues / self.original_variance()

    def eigenvalues_cumulative_ratio(self):
        r"""
        Returns the cumulative ratio between the variance captured by the
        active components and the total amount of variance present on the
        original samples.

        Returns
        -------
        eigenvalues_cumulative_ratio : ``(n_active_components,)`` `ndarray`
            Array of cumulative eigenvalues.
        """
        return np.cumsum(self.eigenvalues_ratio())

    def _total_eigenvalues_cumulative_ratio(self):
        r"""
        Returns the cumulative ratio between the variance captured by the
        active components and the total amount of variance present on the
        original samples.

        Returns
        -------
        total_eigenvalues_cumulative_ratio : ``(n_active_components,)`` `ndarray`
            Array of total cumulative eigenvalues.
        """
        return np.cumsum(self._total_eigenvalues_ratio())

    def noise_variance(self):
        r"""
        Returns the average variance captured by the inactive components,
        i.e. the sample noise assumed in a Probabilistic PCA formulation.

        If all components are active, then ``noise_variance == 0.0``.

        Returns
        -------
        noise_variance : `float`
            The mean variance of the inactive components.
        """
        if self.n_active_components == self.n_components:
            if self._trimmed_eigenvalues.size != 0:
                noise_variance = self._trimmed_eigenvalues.mean()
            else:
                noise_variance = 0.0
        else:
            noise_variance = np.hstack(
                (self._eigenvalues[self.n_active_components:],
                 self._trimmed_eigenvalues)).mean()
        return noise_variance

    def noise_variance_ratio(self):
        r"""
        Returns the ratio between the noise variance and the total amount of
        variance present on the original samples.

        Returns
        -------
        noise_variance_ratio : `float`
            The ratio between the noise variance and the variance present
            in the original samples.
        """
        return self.noise_variance() / self.original_variance()

    def inverse_noise_variance(self):
        r"""
        Returns the inverse of the noise variance.

        Returns
        -------
        inverse_noise_variance : `float`
            Inverse of the noise variance.

        Raises
        ------
        ValueError
            If ``noise_variance() == 0``
        """
        noise_variance = self.noise_variance()
        if np.allclose(noise_variance, 0):
            raise ValueError("noise variance is effectively 0 - "
                             "cannot take the inverse")
        return 1.0 / noise_variance

    def component(self, index, with_mean=True, scale=1.0):
        r"""
        A particular component of the model, in vectorized form.

        Parameters
        ----------
        index : `int`
            The component that is to be returned
        with_mean: `bool`, optional
            If ``True``, the component will be blended with the mean vector
            before being returned. If not, the component is returned on it's
            own.
        scale : `float`, optional
            A scale factor that should be applied to the component. Only
            valid in the case where with_mean is ``True``. The scale is applied
            in units of standard deviations (so a scale of ``1.0``
            `with_mean` visualizes the mean plus ``1`` std. dev of the component
            in question).

        Returns
        -------
        component_vector : ``(n_features,)`` `ndarray`
            The component vector of the given index.
        """
        if with_mean:
            # on PCA, scale is in units of std. deviations...
            scaled_eigval = scale * np.sqrt(self.eigenvalues[index])
            return (scaled_eigval * self.components[index]) + self._mean
        else:
            return self.components[index]

    def instance_vectors(self, weights, normalized_weights=False):
        """
        Creates new vectorized instances of the model using the first
        components in a particular weighting.

        Parameters
        ----------
        weights : ``(n_vectors, n_weights)`` `ndarray` or `list` of `lists`
            The weightings for the first `n_weights` components that
            should be used per instance that is to be produced

            ``weights[i, j]`` is the linear contribution of the j'th
            principal component to the i'th instance vector produced. Note
            that if ``n_weights < n_components``, only the first ``n_weight``
            components are used in the reconstruction (i.e. unspecified
            weights are implicitly ``0``).
        normalized_weights : `bool`, optional
            If ``True``, the weights are assumed to be normalized w.r.t the
            eigenvalues. This can be easier to create unique instances by
            making the weights more interpretable.

        Returns
        -------
        vectors : ``(n_vectors, n_features)`` `ndarray`
            The instance vectors for the weighting provided.

        Raises
        ------
        ValueError
            If n_weights > n_components
        """
        weights = np.asarray(weights)  # if eg a list is provided
        n_instances, n_weights = weights.shape
        if n_weights > self.n_active_components:
            raise ValueError(
                "Number of weightings cannot be greater than {}".format(
                    self.n_active_components))
        else:
            full_weights = np.zeros((n_instances, self.n_active_components),
                                    dtype=self._components.dtype)
            full_weights[..., :n_weights] = weights
            weights = full_weights

        if normalized_weights:
            # If the weights were normalized, then they are all relative to
            # to the scale of the eigenvalues and thus must be multiplied by
            # the sqrt of the eigenvalues.
            weights *= self.eigenvalues ** 0.5
        return self._instance_vectors_for_full_weights(weights)

    def instance(self, weights, normalized_weights=False):
        r"""
        Creates a new vector instance of the model by weighting together the
        components.

        Parameters
        ----------
        weights : ``(n_weights,)`` `ndarray` or `list`
            The weightings for the first `n_weights` components that should be
            used.

            ``weights[j]`` is the linear contribution of the j'th principal
            component to the instance vector.
        normalized_weights : `bool`, optional
            If ``True``, the weights are assumed to be normalized w.r.t the
            eigenvalues. This can be easier to create unique instances by
            making the weights more interpretable.

        Returns
        -------
        vector : ``(n_features,)`` `ndarray`
            The instance vector for the weighting provided.
        """
        weights = np.asarray(weights)
        return self.instance_vectors(
            weights[None, :], normalized_weights=normalized_weights).flatten()

    def trim_components(self, n_components=None):
        r"""
        Permanently trims the components down to a certain amount. The number of
        active components will be automatically reset to this particular value.

        This will reduce `self.n_components` down to `n_components`
        (if ``None``, `self.n_active_components` will be used), freeing up
        memory in the process.

        Once the model is trimmed, the trimmed components cannot be recovered.

        Parameters
        ----------
        n_components: `int` >= ``1`` or `float` > ``0.0`` or ``None``, optional
            The number of components that are kept or else the amount (ratio)
            of variance that is kept. If ``None``, `self.n_active_components` is
            used.

        Notes
        -----
        In case `n_components` is greater than the total number of components or
        greater than the amount of variance currently kept, this method does
        not perform any action.
        """
        if n_components is None:
            # by default trim using the current n_active_components
            n_components = self.n_active_components
        # set self.n_active_components to n_components
        self.n_active_components = n_components

        if self.n_active_components < self.n_components:
            # Just stored so that we can fit < 80 chars
            nac = self.n_active_components
            # set self.n_components to n_components. We have to copy to ensure
            # that the data is actually removed, otherwise a view is returned
            self._components = self._components[:nac].copy()
            # store the eigenvalues associated to the discarded components
            self._trimmed_eigenvalues = np.hstack((
                self._trimmed_eigenvalues,
                self._eigenvalues[self.n_active_components:]))
            # make sure that the eigenvalues are trimmed too
            self._eigenvalues = self._eigenvalues[:nac].copy()

    def project_whitened(self, vector_instance):
        """
        Projects the `vector_instance` onto the whitened components,
        retrieving the whitened linear weightings.

        Parameters
        ----------
        vector_instance : ``(n_features,)`` `ndarray`
            A novel vector.

        Returns
        -------
        projected : ``(n_features,)`` `ndarray`
            A vector of whitened linear weightings
        """
        whitened_components = self.whitened_components()
        return np.dot(vector_instance, whitened_components.T)

    def orthonormalize_against_inplace(self, linear_model):
        r"""
        Enforces that the union of this model's components and another are
        both mutually orthonormal.

        Note that the model passed in is guaranteed to not have it's number
        of available components changed. This model, however, may loose some
        dimensionality due to reaching a degenerate state.

        The removed components will always be trimmed from the end of
        components (i.e. the components which capture the least variance).
        If trimming is performed, `n_components` and `n_available_components`
        would be altered - see :meth:`trim_components` for details.

        Parameters
        ----------
        linear_model : :map:`LinearModel`
            A second linear model to orthonormalize this against.
        """
        # take the QR decomposition of the model components
        Q = (np.linalg.qr(np.hstack((linear_model._components.T,
                                     self._components.T)))[0]).T
        # the model passed to us went first, so all it's components will
        # survive. Pull them off, and update the other model.
        linear_model.components = Q[:linear_model.n_components, :]
        # it's possible that all of our components didn't survive due to
        # degeneracy. We need to trim our components down before replacing
        # them to ensure the number of components is consistent (otherwise
        # the components setter will complain at us)
        n_available_components = Q.shape[0] - linear_model.n_components
        if n_available_components < self.n_components:
            # oh dear, we've lost some components from the end of our model.
            if self.n_active_components < n_available_components:
                # save the current number of active components
                n_active_components = self.n_active_components
            else:
                # save the current number of available components
                n_active_components = n_available_components
            # call trim_components to update our state.
            self.trim_components(n_components=n_available_components)
            if n_active_components < n_available_components:
                # reset the number of active components
                self.n_active_components = n_active_components

        # now we can set our own components with the updated orthogonal ones
        self.components = Q[linear_model.n_components:, :]

    def increment(self, data, n_samples=None, forgetting_factor=1.0,
                  verbose=False):
        r"""
        Update the eigenvectors, eigenvalues and mean vector of this model
        by performing incremental PCA on the given samples.

        Parameters
        ----------
        samples : `list` of :map:`Vectorizable`
            List of new samples to update the model from.
        n_samples : `int`, optional
            If provided then ``samples``  must be an iterator that yields
            ``n_samples``. If not provided then samples has to be a
            list (so we know how large the data matrix needs to be).
        forgetting_factor : ``[0.0, 1.0]`` `float`, optional
            Forgetting factor that weights the relative contribution of new
            samples vs old samples. If 1.0, all samples are weighted equally
            and, hence, the results is the exact same as performing batch
            PCA on the concatenated list of old and new simples. If <1.0,
            more emphasis is put on the new samples. See [1] for details.

        References
        ----------
        .. [1] David Ross, Jongwoo Lim, Ruei-Sung Lin, Ming-Hsuan Yang.
           "Incremental Learning for Robust Visual Tracking". IJCV, 2007.
        """
        data, n_new_samples = self._data_to_matrix(data, n_samples)

        # compute incremental pca
        e_vectors, e_values, m_vector = ipca(
            data, self._components, self._eigenvalues, self.n_samples,
            m_a=self._mean, f=forgetting_factor)

        # if the number of active components is the same as the total number
        # of components so it will be after this method is executed
        reset = (self.n_active_components == self.n_components)

        # update mean, components, eigenvalues and number of samples
        self._mean = m_vector
        self._components = e_vectors
        self._eigenvalues = e_values
        self.n_samples += n_new_samples

        # reset the number of active components to the total number of
        # components
        if reset:
            self.n_active_components = self.n_components

    def plot_eigenvalues(self, figure_id=None, new_figure=False,
                         render_lines=True, line_colour='b', line_style='-',
                         line_width=2, render_markers=True, marker_style='o',
                         marker_size=6, marker_face_colour='b',
                         marker_edge_colour='k', marker_edge_width=1.,
                         render_axes=True, axes_font_name='sans-serif',
                         axes_font_size=10, axes_font_style='normal',
                         axes_font_weight='normal', figure_size=(10, 6),
                         render_grid=True, grid_line_style='--',
                         grid_line_width=0.5):
        r"""
        Plot of the eigenvalues.

        Parameters
        ----------
        figure_id : `object`, optional
            The id of the figure to be used.
        new_figure : `bool`, optional
            If ``True``, a new figure is created.
        render_lines : `bool`, optional
            If ``True``, the line will be rendered.
        line_colour : See Below, optional
            The colour of the lines.
            Example options ::

                {``r``, ``g``, ``b``, ``c``, ``m``, ``k``, ``w``}
                or 
                ``(3, )`` `ndarray`
                or
                `list` of length ``3``

        line_style : {``-``, ``--``, ``-.``, ``:``}, optional
            The style of the lines.
        line_width : `float`, optional
            The width of the lines.
        render_markers : `bool`, optional
            If ``True``, the markers will be rendered.
        marker_style : See Below, optional
            The style of the markers.
            Example options ::

                {``.``, ``,``, ``o``, ``v``, ``^``, ``<``, ``>``, ``+``,
                 ``x``, ``D``, ``d``, ``s``, ``p``, ``*``, ``h``, ``H``,
                 ``1``, ``2``, ``3``, ``4``, ``8``}

        marker_size : `int`, optional
            The size of the markers in points.
        marker_face_colour : See Below, optional
            The face (filling) colour of the markers.
            Example options ::

                {``r``, ``g``, ``b``, ``c``, ``m``, ``k``, ``w``}
                or 
                ``(3, )`` `ndarray`
                or
                `list` of length ``3``

        marker_edge_colour : See Below, optional
            The edge colour of the markers.
            Example options ::

                {``r``, ``g``, ``b``, ``c``, ``m``, ``k``, ``w``}
                or 
                ``(3, )`` `ndarray`
                or
                `list` of length ``3``

        marker_edge_width : `float`, optional
            The width of the markers' edge.
        render_axes : `bool`, optional
            If ``True``, the axes will be rendered.
        axes_font_name : See Below, optional
            The font of the axes.
            Example options ::

                {``serif``, ``sans-serif``, ``cursive``, ``fantasy``,
                 ``monospace``}

        axes_font_size : `int`, optional
            The font size of the axes.
        axes_font_style : {``normal``, ``italic``, ``oblique``}, optional
            The font style of the axes.
        axes_font_weight : See Below, optional
            The font weight of the axes.
            Example options ::

                {``ultralight``, ``light``, ``normal``, ``regular``,
                 ``book``, ``medium``, ``roman``, ``semibold``,
                 ``demibold``, ``demi``, ``bold``, ``heavy``,
                 ``extra bold``, ``black``}

        figure_size : (`float`, `float`) or ``None``, optional
            The size of the figure in inches.
        render_grid : `bool`, optional
            If ``True``, the grid will be rendered.
        grid_line_style : {``-``, ``--``, ``-.``, ``:``}, optional
            The style of the grid lines.
        grid_line_width : `float`, optional
            The width of the grid lines.

        Returns
        -------
        viewer : :map:`MatplotlibRenderer`
            The viewer object.
        """
        from menpo.visualize import plot_curve
        return plot_curve(
            range(self.n_active_components), [self.eigenvalues],
            figure_id=figure_id, new_figure=new_figure, legend_entries=None,
            title='Eigenvalues', x_label='Component Number',
            y_label='Eigenvalue',
            axes_x_limits=[0, self.n_active_components - 1],
            axes_y_limits=None, axes_x_ticks=None, axes_y_ticks=None,
            render_lines=render_lines, line_colour=line_colour,
            line_style=line_style, line_width=line_width,
            render_markers=render_markers, marker_style=marker_style,
            marker_size=marker_size, marker_face_colour=marker_face_colour,
            marker_edge_colour=marker_edge_colour,
            marker_edge_width=marker_edge_width, render_legend=False,
            render_axes=render_axes, axes_font_name=axes_font_name,
            axes_font_size=axes_font_size, axes_font_style=axes_font_style,
            axes_font_weight=axes_font_weight, figure_size=figure_size,
            render_grid=render_grid, grid_line_style=grid_line_style,
            grid_line_width=grid_line_width)

    def plot_eigenvalues_widget(self, figure_size=(10, 6), style='coloured'):
        r"""
        Plot of the eigenvalues using an interactive widget.

        Parameters
        ----------
        figure_size : (`float`, `float`) or ``None``, optional
            The size of the figure in inches.
        style : {``'coloured'``, ``'minimal'``}, optional
            If ``'coloured'``, then the style of the widget will be coloured. If
            ``minimal``, then the style is simple using black and white colours.
        """
        try:
            from menpowidgets import plot_graph
        except:
            from menpo.visualize.base import MenpowidgetsMissingError
            raise MenpowidgetsMissingError()
        plot_graph(x_axis=range(self.n_active_components),
                   y_axis=[self.eigenvalues], legend_entries=['Eigenvalues'],
                   figure_size=figure_size, style=style)

    def plot_eigenvalues_ratio(self, figure_id=None, new_figure=False,
                               render_lines=True, line_colour='b',
                               line_style='-', line_width=2,
                               render_markers=True, marker_style='o',
                               marker_size=6, marker_face_colour='b',
                               marker_edge_colour='k', marker_edge_width=1.,
                               render_axes=True, axes_font_name='sans-serif',
                               axes_font_size=10, axes_font_style='normal',
                               axes_font_weight='normal', figure_size=(10, 6),
                               render_grid=True, grid_line_style='--',
                               grid_line_width=0.5):
        r"""
        Plot of the variance ratio captured by the eigenvalues.

        Parameters
        ----------
        figure_id : `object`, optional
            The id of the figure to be used.
        new_figure : `bool`, optional
            If ``True``, a new figure is created.
        render_lines : `bool`, optional
            If ``True``, the line will be rendered.
        line_colour : See Below, optional
            The colour of the lines.
            Example options ::

                {``r``, ``g``, ``b``, ``c``, ``m``, ``k``, ``w``}
                or 
                ``(3, )`` `ndarray`
                or
                `list` of length ``3``

        line_style : {``-``, ``--``, ``-.``, ``:``}, optional
            The style of the lines.
        line_width : `float`, optional
            The width of the lines.
        render_markers : `bool`, optional
            If ``True``, the markers will be rendered.
        marker_style : See Below, optional
            The style of the markers.
            Example options ::

                {``.``, ``,``, ``o``, ``v``, ``^``, ``<``, ``>``, ``+``,
                 ``x``, ``D``, ``d``, ``s``, ``p``, ``*``, ``h``, ``H``,
                 ``1``, ``2``, ``3``, ``4``, ``8``}

        marker_size : `int`, optional
            The size of the markers in points.
        marker_face_colour : See Below, optional
            The face (filling) colour of the markers.
            Example options ::

                {``r``, ``g``, ``b``, ``c``, ``m``, ``k``, ``w``}
                or 
                ``(3, )`` `ndarray`
                or
                `list` of length ``3``

        marker_edge_colour : See Below, optional
            The edge colour of the markers.
            Example options ::

                {``r``, ``g``, ``b``, ``c``, ``m``, ``k``, ``w``}
                or 
                ``(3, )`` `ndarray`
                or
                `list` of length ``3``

        marker_edge_width : `float`, optional
            The width of the markers' edge.
        render_axes : `bool`, optional
            If ``True``, the axes will be rendered.
        axes_font_name : See Below, optional
            The font of the axes.
            Example options ::

                {``serif``, ``sans-serif``, ``cursive``, ``fantasy``,
                 ``monospace``}

        axes_font_size : `int`, optional
            The font size of the axes.
        axes_font_style : {``normal``, ``italic``, ``oblique``}, optional
            The font style of the axes.
        axes_font_weight : See Below, optional
            The font weight of the axes.
            Example options ::

                {``ultralight``, ``light``, ``normal``, ``regular``,
                 ``book``, ``medium``, ``roman``, ``semibold``,
                 ``demibold``, ``demi``, ``bold``, ``heavy``,
                 ``extra bold``, ``black``}

        figure_size : (`float`, `float`) or `None`, optional
            The size of the figure in inches.
        render_grid : `bool`, optional
            If ``True``, the grid will be rendered.
        grid_line_style : {``-``, ``--``, ``-.``, ``:``}, optional
            The style of the grid lines.
        grid_line_width : `float`, optional
            The width of the grid lines.

        Returns
        -------
        viewer : :map:`MatplotlibRenderer`
            The viewer object.
        """
        from menpo.visualize import plot_curve
        return plot_curve(
            range(self.n_active_components), [self.eigenvalues_ratio()],
            figure_id=figure_id, new_figure=new_figure, legend_entries=None,
            title='Variance Ratio of Eigenvalues', x_label='Component Number',
            y_label='Variance Ratio',
            axes_x_limits=[0, self.n_active_components - 1],
            axes_y_limits=None, axes_x_ticks=None, axes_y_ticks=None,
            render_lines=render_lines, line_colour=line_colour,
            line_style=line_style, line_width=line_width,
            render_markers=render_markers, marker_style=marker_style,
            marker_size=marker_size, marker_face_colour=marker_face_colour,
            marker_edge_colour=marker_edge_colour,
            marker_edge_width=marker_edge_width, render_legend=False,
            render_axes=render_axes, axes_font_name=axes_font_name,
            axes_font_size=axes_font_size, axes_font_style=axes_font_style,
            axes_font_weight=axes_font_weight, figure_size=figure_size,
            render_grid=render_grid, grid_line_style=grid_line_style,
            grid_line_width=grid_line_width)

    def plot_eigenvalues_ratio_widget(self, figure_size=(10, 6),
                                      style='coloured'):
        r"""
        Plot of the variance ratio captured by the eigenvalues using an
        interactive widget.

        Parameters
        ----------
        figure_size : (`float`, `float`) or ``None``, optional
            The size of the figure in inches.
        style : {``'coloured'``, ``'minimal'``}, optional
            If ``'coloured'``, then the style of the widget will be coloured. If
            ``minimal``, then the style is simple using black and white colours.
        """
        try:
            from menpowidgets import plot_graph
        except:
            from menpo.visualize.base import MenpowidgetsMissingError
            raise MenpowidgetsMissingError()
        plot_graph(x_axis=range(self.n_active_components),
                   y_axis=[self.eigenvalues_ratio()],
                   legend_entries=['Eigenvalues ratio'],
                   figure_size=figure_size, style=style)

    def plot_eigenvalues_cumulative_ratio(self, figure_id=None,
                                          new_figure=False, render_lines=True,
                                          line_colour='b', line_style='-',
                                          line_width=2, render_markers=True,
                                          marker_style='o', marker_size=6,
                                          marker_face_colour='b',
                                          marker_edge_colour='k',
                                          marker_edge_width=1.,
                                          render_axes=True,
                                          axes_font_name='sans-serif',
                                          axes_font_size=10,
                                          axes_font_style='normal',
                                          axes_font_weight='normal',
                                          figure_size=(10, 6), render_grid=True,
                                          grid_line_style='--',
                                          grid_line_width=0.5):
        r"""
        Plot of the cumulative variance ratio captured by the eigenvalues.

        Parameters
        ----------
        figure_id : `object`, optional
            The id of the figure to be used.
        new_figure : `bool`, optional
            If ``True``, a new figure is created.
        render_lines : `bool`, optional
            If ``True``, the line will be rendered.
        line_colour : See Below, optional
            The colour of the lines.
            Example options ::

                {``r``, ``g``, ``b``, ``c``, ``m``, ``k``, ``w``}
                or 
                ``(3, )`` `ndarray`
                or
                `list` of length ``3``

        line_style : {``-``, ``--``, ``-.``, ``:``}, optional
            The style of the lines.
        line_width : `float`, optional
            The width of the lines.
        render_markers : `bool`, optional
            If ``True``, the markers will be rendered.
        marker_style : See Below, optional
            The style of the markers.
            Example options ::

                {``.``, ``,``, ``o``, ``v``, ``^``, ``<``, ``>``, ``+``,
                 ``x``, ``D``, ``d``, ``s``, ``p``, ``*``, ``h``, ``H``,
                 ``1``, ``2``, ``3``, ``4``, ``8``}

        marker_size : `int`, optional
            The size of the markers in points.
        marker_face_colour : See Below, optional
            The face (filling) colour of the markers.
            Example options ::

                {``r``, ``g``, ``b``, ``c``, ``m``, ``k``, ``w``}
                or 
                ``(3, )`` `ndarray`
                or
                `list` of length ``3``

        marker_edge_colour : See Below, optional
            The edge colour of the markers.
            Example options ::

                {``r``, ``g``, ``b``, ``c``, ``m``, ``k``, ``w``}
                or 
                ``(3, )`` `ndarray`
                or
                `list` of length ``3``

        marker_edge_width : `float`, optional
            The width of the markers' edge.
        render_axes : `bool`, optional
            If ``True``, the axes will be rendered.
        axes_font_name : See Below, optional
            The font of the axes.
            Example options ::

                {``serif``, ``sans-serif``, ``cursive``, ``fantasy``,
                 ``monospace``}

        axes_font_size : `int`, optional
            The font size of the axes.
        axes_font_style : {``normal``, ``italic``, ``oblique``}, optional
            The font style of the axes.
        axes_font_weight : See Below, optional
            The font weight of the axes.
            Example options ::

                {``ultralight``, ``light``, ``normal``, ``regular``,
                 ``book``, ``medium``, ``roman``, ``semibold``,
                 ``demibold``, ``demi``, ``bold``, ``heavy``,
                 ``extra bold``, ``black``}

        figure_size : (`float`, `float`) or `None`, optional
            The size of the figure in inches.
        render_grid : `bool`, optional
            If ``True``, the grid will be rendered.
        grid_line_style : {``-``, ``--``, ``-.``, ``:``}, optional
            The style of the grid lines.
        grid_line_width : `float`, optional
            The width of the grid lines.

        Returns
        -------
        viewer : :map:`MatplotlibRenderer`
            The viewer object.
        """
        from menpo.visualize import plot_curve
        return plot_curve(
            range(self.n_active_components),
            [self.eigenvalues_cumulative_ratio()], figure_id=figure_id,
            new_figure=new_figure, legend_entries=None,
            title='Cumulative Variance Ratio of Eigenvalues',
            x_label='Component Number', y_label='Cumulative Variance Ratio',
            axes_x_limits=[0, self.n_active_components - 1],
            axes_y_limits=None, axes_x_ticks=None, axes_y_ticks=None,
            render_lines=render_lines, line_colour=line_colour,
            line_style=line_style, line_width=line_width,
            render_markers=render_markers, marker_style=marker_style,
            marker_size=marker_size, marker_face_colour=marker_face_colour,
            marker_edge_colour=marker_edge_colour,
            marker_edge_width=marker_edge_width, render_legend=False,
            render_axes=render_axes, axes_font_name=axes_font_name,
            axes_font_size=axes_font_size, axes_font_style=axes_font_style,
            axes_font_weight=axes_font_weight, figure_size=figure_size,
            render_grid=render_grid, grid_line_style=grid_line_style,
            grid_line_width=grid_line_width)

    def plot_eigenvalues_cumulative_ratio_widget(self, figure_size=(10, 6),
                                                 style='coloured'):
        r"""
        Plot of the cumulative variance ratio captured by the eigenvalues using
        an interactive widget.

        Parameters
        ----------
        figure_size : (`float`, `float`) or ``None``, optional
            The size of the figure in inches.
        style : {``'coloured'``, ``'minimal'``}, optional
            If ``'coloured'``, then the style of the widget will be coloured. If
            ``minimal``, then the style is simple using black and white colours.
        """
        try:
            from menpowidgets import plot_graph
        except:
            from menpo.visualize.base import MenpowidgetsMissingError
            raise MenpowidgetsMissingError()
        plot_graph(x_axis=range(self.n_active_components),
                   y_axis=[self.eigenvalues_cumulative_ratio()],
                   legend_entries=['Eigenvalues cumulative ratio'],
                   figure_size=figure_size, style=style)

    def __str__(self):
        str_out = 'PCA Vector Model \n'                      \
                  ' - centred:              {}\n'            \
                  ' - # features:           {}\n'            \
                  ' - # active components:  {}\n'            \
                  ' - kept variance:        {:.2}  {:.1%}\n' \
                  ' - noise variance:       {:.2}  {:.1%}\n' \
                  ' - total # components:   {}\n'            \
                  ' - components shape:     {}\n'.format(
            self.centred,  self.n_features, self.n_active_components,
            self.variance(), self.variance_ratio(), self.noise_variance(),
            self.noise_variance_ratio(), self.n_components,
            self.components.shape)
        return str_out


class PCAModel(VectorizableBackedModel, PCAVectorModel):
    r"""
    A :map:`MeanLinearModel` where components are Principal Components
    and the components are vectorized instances.

    Principal Component Analysis (PCA) by eigenvalue decomposition of the
    data's scatter matrix. For details of the implementation of PCA, see
    :map:`pca`.

    Parameters
    ----------
    samples : `list` or `iterable` of :map:`Vectorizable`
        List or iterable of samples to build the model from.
    centre : `bool`, optional
        When ``True`` (default) PCA is performed after mean centering the data.
        If ``False`` the data is assumed to be centred, and the mean will be
        ``0``.
    n_samples : `int`, optional
        If provided then ``samples``  must be an iterator that yields
        ``n_samples``. If not provided then samples has to be a `list` (so we
        know how large the data matrix needs to be).
    max_n_components : `int`, optional
        The maximum number of components to keep in the model. Any components
        above and beyond this one are discarded.
    inplace : `bool`, optional
        If ``True`` the data matrix is modified in place. Otherwise, the data
        matrix is copied.
    verbose : `bool`, optional
        Whether to print building information or not.
     """

    def __init__(self, samples, centre=True, n_samples=None,
                 max_n_components=None, inplace=True, verbose=False):
        # build a data matrix from all the samples
        data, template = as_matrix(samples, length=n_samples,
                                   return_template=True, verbose=verbose)
        n_samples = data.shape[0]

        PCAVectorModel.__init__(self, data, centre=centre,
                                max_n_components=max_n_components,
                                n_samples=n_samples, inplace=inplace)
        VectorizableBackedModel.__init__(self, template)

    @classmethod
    def init_from_covariance_matrix(cls, C, mean, n_samples, centred=True,
                                    is_inverse=False, max_n_components=None):
        r"""
        Build the Principal Component Analysis (PCA) by eigenvalue
        decomposition of the provided covariance/scatter matrix. For details
        of the implementation of PCA, see :map:`pcacov`.

        Parameters
        ----------
        C : ``(n_features, n_features)`` `ndarray` or `scipy.sparse`
            The Covariance/Scatter matrix. If it is a precision matrix (inverse
            covariance), then set `is_inverse=True`.
        mean : :map:`Vectorizable`
            The mean instance. It must be a :map:`Vectorizable` and *not* an
            `ndarray`.
        n_samples : `int`
            The number of samples used to generate the covariance matrix.
        centred : `bool`, optional
            When ``True`` we assume that the data were centered before
            computing the covariance matrix.
        is_inverse : `bool`, optional
            It ``True``, then it is assumed that `C` is a precision matrix (
            inverse covariance). Thus, the eigenvalues will be inverted. If
            ``False``, then it is assumed that `C` is a covariance matrix.
        max_n_components : `int`, optional
            The maximum number of components to keep in the model. Any
            components above and beyond this one are discarded.
        """
        # Create new pca instance
        self_model = PCAVectorModel.__new__(cls)
        self_model.n_samples = n_samples

        # Compute pca on covariance
        e_vectors, e_values = pcacov(C, is_inverse=is_inverse)

        # The call to __init__ of MeanLinearModel is done in here
        self_model._constructor_helper(eigenvalues=e_values,
                                       eigenvectors=e_vectors,
                                       mean=mean.as_vector(),
                                       centred=centred,
                                       max_n_components=max_n_components)
        VectorizableBackedModel.__init__(self_model, mean)
        return self_model

    @classmethod
    def init_from_components(cls, components, eigenvalues, mean, n_samples,
                             centred, max_n_components=None):
        r"""
        Build the Principal Component Analysis (PCA) using the provided
        components (eigenvectors) and eigenvalues.

        Parameters
        ----------
        components : ``(n_components, n_features)`` `ndarray`
            The eigenvectors to be used.
        eigenvalues : ``(n_components, )`` `ndarray`
            The corresponding eigenvalues.
        mean : :map:`Vectorizable`
            The mean instance. It must be a :map:`Vectorizable` and *not* an
            `ndarray`.
        n_samples : `int`
            The number of samples used to generate the eigenvectors.
        centred : `bool`, optional
            When ``True`` we assume that the data were centered before
            computing the eigenvectors.
        max_n_components : `int`, optional
            The maximum number of components to keep in the model. Any
            components above and beyond this one are discarded.
        """
        # Create new pca instance
        self_model = PCAVectorModel.__new__(cls)
        self_model.n_samples = n_samples

        # The call to __init__ of MeanLinearModel is done in here
        self_model._constructor_helper(
            eigenvalues=eigenvalues, eigenvectors=components,
            mean=mean.as_vector(), centred=centred,
            max_n_components=max_n_components)
        VectorizableBackedModel.__init__(self_model, mean)
        return self_model

    def mean(self):
        r"""
        Return the mean of the model.

        :type: :map:`Vectorizable`
        """
        return self.template_instance.from_vector(self._mean)

    @property
    def mean_vector(self):
        r"""
        Return the mean of the model as a 1D vector.

        :type: `ndarray`
        """
        return self._mean

    @doc_inherit(name='project_out')
    def project_out_vector(self, instance_vector):
        return PCAVectorModel.project_out(self, instance_vector)

    @doc_inherit(name='reconstruct')
    def reconstruct_vector(self, instance_vector):
        return PCAVectorModel.reconstruct(self, instance_vector)

    @doc_inherit(name='project')
    def project_vector(self, instance_vector):
        return PCAVectorModel.project(self, instance_vector)

    @doc_inherit(name='instance')
    def instance_vector(self, weights, normalized_weights=False):
        return PCAVectorModel.instance(self, weights,
                                       normalized_weights=normalized_weights)

    @doc_inherit(name='component')
    def component_vector(self, index, with_mean=True, scale=1.0):
        return PCAVectorModel.component(self, index, with_mean=with_mean,
                                        scale=scale)

    @doc_inherit(name='project_whitened')
    def project_whitened_vector(self, vector_instance):
        return PCAVectorModel.project_whitened(self, vector_instance)

    def component(self, index, with_mean=True, scale=1.0):
        r"""
        Return a particular component of the linear model.

        Parameters
        ----------
        index : `int`
            The component that is to be returned
        with_mean: `bool`, optional
            If ``True``, the component will be blended with the mean vector
            before being returned. If not, the component is returned on it's
            own.
        scale : `float`, optional
            A scale factor that should be applied to the component. Only
            valid in the case where ``with_mean == True``. See
            :meth:`component_vector` for how this scale factor is interpreted.

        Returns
        -------
        component : `type(self.template_instance)`
            The requested component instance.
        """
        return self.template_instance.from_vector(self.component_vector(
            index, with_mean=with_mean, scale=scale))

    def instance(self, weights, normalized_weights=False):
        """
        Creates a new instance of the model using the first ``len(weights)``
        components.

        Parameters
        ----------
        weights : ``(n_weights,)`` `ndarray` or `list`
            ``weights[i]`` is the linear contribution of the i'th component
            to the instance vector.
        normalized_weights : `bool`, optional
            If ``True``, the weights are assumed to be normalized w.r.t the
            eigenvalues. This can be easier to create unique instances by
            making the weights more interpretable.
        Raises
        ------
        ValueError
            If n_weights > n_components

        Returns
        -------
        instance : `type(self.template_instance)`
            An instance of the model.
        """
        v = self.instance_vector(weights, normalized_weights=normalized_weights)
        return self.template_instance.from_vector(v)

    def project_whitened(self, instance):
        """
        Projects the `instance` onto the whitened components, retrieving the 
        whitened linear weightings.

        Parameters
        ----------
        instance : :map:`Vectorizable`
            A novel instance.

        Returns
        -------
        projected : (n_components,)
            A vector of whitened linear weightings
        """
        return self.project_whitened_vector(instance.as_vector())

    def increment(self, samples, n_samples=None, forgetting_factor=1.0,
                  verbose=False):
        r"""
        Update the eigenvectors, eigenvalues and mean vector of this model
        by performing incremental PCA on the given samples.

        Parameters
        ----------
        samples : `list` of :map:`Vectorizable`
            List of new samples to update the model from.
        n_samples : `int`, optional
            If provided then ``samples``  must be an iterator that yields
            ``n_samples``. If not provided then samples has to be a
            list (so we know how large the data matrix needs to be).
        forgetting_factor : ``[0.0, 1.0]`` `float`, optional
            Forgetting factor that weights the relative contribution of new
            samples vs old samples. If 1.0, all samples are weighted equally
            and, hence, the results is the exact same as performing batch
            PCA on the concatenated list of old and new simples. If <1.0,
            more emphasis is put on the new samples. See [1] for details.

        References
        ----------
        .. [1] David Ross, Jongwoo Lim, Ruei-Sung Lin, Ming-Hsuan Yang.
           "Incremental Learning for Robust Visual Tracking". IJCV, 2007.
        """
        # build a data matrix from the new samples
        data = as_matrix(samples, length=n_samples, verbose=verbose)
        n_new_samples = data.shape[0]
        PCAVectorModel.increment(self, data, n_samples=n_new_samples,
                                 forgetting_factor=forgetting_factor,
                                 verbose=verbose)

    def __str__(self):
        str_out = 'PCA Model \n'                             \
                  ' - instance class:       {}\n'            \
                  ' - centred:              {}\n'            \
                  ' - # features:           {}\n'            \
                  ' - # active components:  {}\n'            \
                  ' - kept variance:        {:.2}  {:.1%}\n' \
                  ' - noise variance:       {:.2}  {:.1%}\n' \
                  ' - total # components:   {}\n'            \
                  ' - components shape:     {}\n'.format(
            name_of_callable(self.template_instance), self.centred,
            self.n_features, self.n_active_components, self.variance(),
            self.variance_ratio(), self.noise_variance(),
            self.noise_variance_ratio(), self.n_components,
            self.components.shape)
        return str_out



from mock import Mock

from menpo.transform import TransformChain, Transform


def transformchain_apply_test():
    mocked_transform = Mock()
    mocked_transform._apply.return_value = 3
    mocked_transform2 = Mock()
    mocked_transform2._apply.return_value = 4
    transforms = [mocked_transform, mocked_transform2]
    tr = TransformChain(transforms)
    result = tr.apply(1)
    assert (result == 4)


def transformchain_composes_inplace_with_test():
    tr = TransformChain([])
    assert (tr.composes_inplace_with == Transform)


def transformchain_compose_before_composes_with_test():
    tr = TransformChain([])
    new_tr = tr.compose_before(Mock(spec=Transform))
    assert (new_tr is not tr)
    assert (len(new_tr.transforms) is 1)


def transformchain_compose_before_inplace_order_test():
    m1 = Mock(spec=Transform)
    m2 = Mock(spec=Transform)
    tr = TransformChain([m1])
    tr.compose_before_inplace(m2)
    assert (tr.transforms[1] is m2)


def transformchain_compose_after_inplace_order_test():
    m1 = Mock(spec=Transform)
    m2 = Mock(spec=Transform)
    tr = TransformChain([m1])
    tr.compose_after_inplace(m2)
    assert (tr.transforms[0] is m2)


def transformchain_compose_before_inplace_composes_with_test():
    tr = TransformChain([])
    ref = tr
    no_return = tr.compose_after_inplace(Mock(spec=Transform))
    assert (no_return is None)
    assert (ref is tr)
    assert (len(tr.transforms) is 1)


def transformchain_compose_after_composes_with_test():
    tr = TransformChain([])
    new_tr = tr.compose_after(Mock(spec=Transform))
    assert (new_tr is not tr)
    assert (len(new_tr.transforms) is 1)


def transformchain_compose_after_inplace_composes_with_test():
    tr = TransformChain([])
    ref = tr
    no_return = tr.compose_after_inplace(Mock(spec=Transform))
    assert (no_return is None)
    assert (ref is tr)
    assert (len(tr.transforms) is 1)



import numpy as np
from scipy.misc import imrotate

from .base import ndfeature


@ndfeature
def sum_channels(pixels, channels=None):
    r"""
    Create the sum of the channels of an image that can be used for
    visualization.

    Parameters
    ----------
    pixels : :map:`Image` or subclass or ``(C, X, Y, ..., Z)`` `ndarray`
        Either the image object itself or an array with the pixels. The first
        dimension is interpreted as channels.
    channels : `list` of `int` or ``None``
        The list of channels to be used. If ``None``, then all the channels are
        employed.
    """
    # if channels is None, then all channels are used
    if channels is None:
        # Not indexing is twice as fast
        sum_image = np.sum(pixels, axis=0)
    else:
        sum_image = np.sum(pixels[channels], axis=0)
    return sum_image.reshape((1,) + sum_image.shape)  # add a channel axis


# TODO: Needs fixing ...
@ndfeature
def glyph(pixels, vectors_block_size=10, use_negative=False, channels=None):
    r"""
    Create the glyph of a feature image that can be used for visualization. If
    `pixels` have negative values, the `use_negative` flag controls whether
    there will be created a glyph of both positive and negative values
    concatenated the one on top of the other.

    Parameters
    ----------
    pixels : :map:`Image` or subclass or ``(C, X, Y, ..., Z)`` `ndarray`
        Either the image object itself or an array with the pixels. The first
        dimension is interpreted as channels.
    vectors_block_size : `int`
        Defines the size of each block with vectors of the glyph image.
    use_negative : `bool`
        Defines whether to take into account possible negative values of
        feature_data.
    channels : `list` of `int` or ``None``
        The list of channels to be used. If ``None``, then all the channels are
        employed.
    """
    # first, choose the appropriate channels
    if channels is None:
        pixels = pixels[:4]
    elif channels != 'all':
        pixels = pixels[channels]
    else:
        pixels = pixels

    # TODO: This is a temporal fix
    # flip axis
    pixels = np.rollaxis(pixels, 0, len(pixels.shape))

    # compute the glyph
    negative_weights = -pixels
    scale = np.maximum(pixels.max(), negative_weights.max())
    pos = _create_feature_glyph(pixels, vectors_block_size)
    pos = pos * 255 / scale
    glyph_image = pos
    if use_negative and pixels.min() < 0:
        neg = _create_feature_glyph(negative_weights, vectors_block_size)
        neg = neg * 255 / scale
        glyph_image = np.concatenate((pos, neg))
    # return as c-contiguous
    return np.ascontiguousarray(glyph_image[None, ...])  # add a channel axis


def _create_feature_glyph(feature, vbs):
    r"""
    Create glyph of feature pixels.

    Parameters
    ----------
    feature : ``(N, D)`` `ndarray`
        The feature pixels to use.
    vbs : `int`
        Defines the size of each block with vectors of the glyph image.
    """
    # vbs = Vector block size
    num_bins = feature.shape[2]
    # construct a "glyph" for each orientation
    block_image_temp = np.zeros((vbs, vbs))
    # Create a vertical line of ones, to be the first vector
    block_image_temp[:, np.round(vbs / 2) - 1:np.round(vbs / 2) + 1] = 1
    block_im = np.zeros((block_image_temp.shape[0],
                         block_image_temp.shape[1],
                         num_bins))
    # First vector as calculated above
    block_im[:, :, 0] = block_image_temp
    # Number of bins rotations to create an 'asterisk' shape
    for i in range(1, num_bins):
        block_im[:, :, i] = imrotate(block_image_temp, -i * vbs)

    # make pictures of positive feature_data by adding up weighted glyphs
    feature[feature < 0] = 0
    glyph_im = np.sum(block_im[None, None, :, :, :] *
                      feature[:, :, None, None, :], axis=-1)
    glyph_im = np.bmat(glyph_im.tolist())
    return glyph_im



import numpy as np

from .base import Homogeneous, HomogFamilyAlignment
from functools import reduce


class Affine(Homogeneous):
    r"""
    Base class for all ``n``-dimensional affine transformations. Provides
    methods to break the transform down into its constituent
    scale/rotation/translation, to view the homogeneous matrix equivalent,
    and to chain this transform with other affine transformations.

    Parameters
    ----------
    h_matrix : ``(n_dims + 1, n_dims + 1)`` `ndarray`
        The homogeneous matrix of the affine transformation.
    copy : `bool`, optional
        If ``False`` avoid copying ``h_matrix`` for performance.
    skip_checks : `bool`, optional
        If ``True`` avoid sanity checks on ``h_matrix`` for performance.
    """
    def __init__(self, h_matrix, copy=True, skip_checks=False):
        Homogeneous.__init__(self, h_matrix, copy=copy,
                             skip_checks=skip_checks)

    @classmethod
    def init_identity(cls, n_dims):
        r"""
        Creates an identity matrix Affine transform.

        Parameters
        ----------
        n_dims : `int`
            The number of dimensions.

        Returns
        -------
        identity : :class:`Affine`
            The identity matrix transform.
        """
        return cls(np.eye(n_dims + 1), copy=False, skip_checks=True)

    @property
    def h_matrix(self):
        r"""
        The homogeneous matrix defining this transform.

        :type: ``(n_dims + 1, n_dims + 1)`` `ndarray`
        """
        return self._h_matrix

    def _set_h_matrix(self, value, copy=True, skip_checks=False):
        r"""
        Updates the `h_matrix`, performing sanity checks.

        Parameters
        ----------
        value : `ndarray`
            The new homogeneous matrix to set
        copy : `bool`, optional
            If ``False`` do not copy the h_matrix. Useful for performance.
        skip_checks : `bool`, optional
            If ``True`` skip sanity checks on the matrix. Useful for performance.
        """
        if not skip_checks:
            shape = value.shape
            if len(shape) != 2 or shape[0] != shape[1]:
                raise ValueError("You need to provide a square homogeneous "
                                 "matrix")
            if self.h_matrix is not None:
                # already have a matrix set! The update better be the same size
                if self.n_dims != shape[0] - 1:
                    raise ValueError("Trying to update the homogeneous "
                                     "matrix to a different dimension")
            if shape[0] - 1 not in [2, 3]:
                raise ValueError("Affine Transforms can only be 2D or 3D")
            if not (np.allclose(value[-1, :-1], 0) and
                    np.allclose(value[-1, -1], 1)):
                raise ValueError("Bottom row must be [0 0 0 1] or [0, 0, 1]")
        if copy:
            value = value.copy()
        self._h_matrix = value

    @property
    def linear_component(self):
        r"""
        The linear component of this affine transform.

        :type: ``(n_dims, n_dims)`` `ndarray`
        """
        return self.h_matrix[:-1, :-1]

    @property
    def translation_component(self):
        r"""
        The translation component of this affine transform.

        :type: ``(n_dims,)`` `ndarray`
        """
        return self.h_matrix[:-1, -1]

    def decompose(self):
        r"""
        Decompose this transform into discrete Affine Transforms.

        Useful for understanding the effect of a complex composite transform.

        Returns
        -------
        transforms : `list` of :map:`DiscreteAffine`
            Equivalent to this affine transform, such that

            .. code-block:: python

                reduce(lambda x, y: x.chain(y), self.decompose()) == self

        """
        from .rotation import Rotation
        from .translation import Translation
        from .scale import Scale
        U, S, V = np.linalg.svd(self.linear_component)
        rotation_2 = Rotation(U)
        rotation_1 = Rotation(V)
        scale = Scale(S)
        translation = Translation(self.translation_component)
        return [rotation_1, scale, rotation_2, translation]

    def _transform_str(self):
        r"""
        A string representation explaining what this affine transform does.
        Has to be implemented by base classes.

        Returns
        -------
        str : `str`
            String representation of transform.
        """
        header = 'Affine decomposing into:'
        list_str = [t._transform_str() for t in self.decompose()]
        return header + reduce(lambda x, y: x + '\n' + '  ' + y, list_str, '  ')

    def _apply(self, x, **kwargs):
        r"""
        Applies this transform to a new set of vectors.

        Parameters
        ----------
        x : ``(N, D)`` `ndarray`
            Array to apply this transform to.

        Returns
        -------
        transformed_x : ``(N, D)`` `ndarray`
            The transformed array.
        """
        return np.dot(x, self.linear_component.T) + self.translation_component

    @property
    def n_parameters(self):
        r"""
        ``n_dims * (n_dims + 1)`` parameters - every element of the matrix but
        the homogeneous part.

        :type: int

        Examples
        --------
        2D Affine: 6 parameters::

            [p1, p3, p5]
            [p2, p4, p6]


        3D Affine: 12 parameters::

            [p1, p4, p7, p10]
            [p2, p5, p8, p11]
            [p3, p6, p9, p12]

        """
        return self.n_dims * (self.n_dims + 1)

    def _as_vector(self):
        r"""
        Return the parameters of the transform as a 1D array. These parameters
        are parametrised as deltas from the identity warp. This does not
        include the homogeneous part of the warp. Note that it flattens using
        Fortran ordering, to stay consistent with Matlab.

        **2D**

        ========= ===========================================
        parameter definition
        ========= ===========================================
        p1        Affine parameter
        p2        Affine parameter
        p3        Affine parameter
        p4        Affine parameter
        p5        Translation in `x`
        p6        Translation in `y`
        ========= ===========================================

        3D and higher transformations follow a similar format to the 2D case.

        Returns
        -------
        params : ``(n_parameters,)`` `ndarray`
            The values that parametrise the transform.
        """
        params = self.h_matrix - np.eye(self.n_dims + 1)
        return params[:self.n_dims, :].ravel(order='F')

    def _from_vector_inplace(self, p):
        r"""
        Updates this Affine in-place from the new parameters. See
        from_vector for details of the parameter format
        """
        h_matrix = None
        if p.shape[0] == 6:  # 2D affine
            h_matrix = np.eye(3)
            h_matrix[:2, :] += p.reshape((2, 3), order='F')
        elif p.shape[0] == 12:  # 3D affine
            h_matrix = np.eye(4)
            h_matrix[:3, :] += p.reshape((3, 4), order='F')
        else:
            ValueError("Only 2D (6 parameters) or 3D (12 parameters) "
                       "homogeneous matrices are supported.")
        self._set_h_matrix(h_matrix, copy=False, skip_checks=True)

    @property
    def composes_inplace_with(self):
        r"""
        :class:`Affine` can swallow composition with any other :class:`Affine`.
        """
        return Affine


class AlignmentAffine(HomogFamilyAlignment, Affine):
    r"""
    Constructs an :class:`Affine` by finding the optimal affine transform to
    align `source` to `target`.

    Parameters
    ----------
    source : :map:`PointCloud`
        The source pointcloud instance used in the alignment
    target : :map:`PointCloud`
        The target pointcloud instance used in the alignment

    Notes
    -----
    We want to find the optimal transform M which satisfies :math:`M a = b`
    where :math:`a` and :math:`b` are the `source` and `target` homogeneous
    vectors respectively. ::

       (M a)' = b'
       a' M' = b'
       a a' M' = a b'

    `a a'` is of shape `(n_dim + 1, n_dim + 1)` and so can be inverted
    to solve for `M`.

    This approach is the analytical linear least squares solution to the
    problem at hand. It will have a solution as long as `(a a')` is
    non-singular, which generally means at least 2 corresponding points are
    required.
    """
    def __init__(self, source, target):
        # first, initialize the alignment
        HomogFamilyAlignment.__init__(self, source, target)
        # now, the Affine
        optimal_h = self._build_alignment_h_matrix(source, target)
        Affine.__init__(self, optimal_h, copy=False, skip_checks=True)

    @staticmethod
    def _build_alignment_h_matrix(source, target):
        r"""
        Returns the optimal alignment of `source` to `target`.

        Parameters
        ----------
        source : :map:`PointCloud`
            The source pointcloud instance used in the alignment
        target : :map:`PointCloud`
            The target pointcloud instance used in the alignment
        """
        a = source.h_points()
        b = target.h_points()
        return np.linalg.solve(np.dot(a, a.T), np.dot(a, b.T)).T

    def _set_h_matrix(self, value, copy=True, skip_checks=False):
        r"""
        Updates ``h_matrix``, optionally performing sanity checks.

        .. note::

            Updating the ``h_matrix`` on an :map:`AlignmentAffine`
            triggers a sync of the target.

        Note that it won't always be possible to manually specify the
        ``h_matrix`` through this method, specifically if changing the
        ``h_matrix`` could change the nature of the transform. See
        :attr:`h_matrix_is_mutable` for how you can discover if the
        ``h_matrix`` is allowed to be set for a given class.

        Parameters
        ----------
        value : `ndarray`
            The new homogeneous matrix to set
        copy : `bool`, optional
            If ``False`` do not copy the h_matrix. Useful for performance.
        skip_checks : `bool`, optional
            If ``True`` skip checking. Useful for performance.

        Raises
        ------
        NotImplementedError
            If :attr:`h_matrix_is_mutable` returns ``False``.
        """
        Affine._set_h_matrix(self, value, copy=copy, skip_checks=skip_checks)
        # now update the state
        self._sync_target_from_state()

    def _sync_state_from_target(self):
        optimal_h = self._build_alignment_h_matrix(self.source, self.target)
        # Use the pure Affine setter (so we don't get syncing)
        # We know the resulting affine is correct so skip the checks
        Affine._set_h_matrix(self, optimal_h, copy=False, skip_checks=True)

    def as_non_alignment(self):
        r"""
        Returns a copy of this :map:`Affine` without its alignment nature.

        Returns
        -------
        transform : :map:`Affine`
            A version of this affine with the same transform behavior but
            without the alignment logic.
        """
        return Affine(self.h_matrix, skip_checks=True)


class DiscreteAffine(object):
    r"""
    A discrete Affine transform operation (such as a :meth:`Scale`,
    :class:`Translation` or :meth:`Rotation`). Has to be invertable. Make sure
    you inherit from :class:`DiscreteAffine` first, for optimal
    `decompose()` behavior.
    """

    def decompose(self):
        r"""
        A :class:`DiscreteAffine` is already maximally decomposed -
        return a copy of self in a `list`.

        Returns
        -------
        transform : :class:`DiscreteAffine`
            Deep copy of `self`.
        """
        return [self.copy()]



import os
import numpy as np
import unittest
from numpy.testing import assert_allclose
from menpo.shape import PointCloud, PointUndirectedGraph
from menpo.image import Image
from menpo.image.rasterize import rasterize_landmarks_2d


centre = PointCloud([[4.5, 4.5]])
line = PointUndirectedGraph(np.array([[2, 4.5], [8, 4.5]]),
                            adjacency_matrix=np.array([[0, 1], [1, 0]]))


@unittest.skipIf(os.environ.get('TRAVIS'), 'Skipping due to lack of DISPLAY on Travis')
def test_rasterize_matplotlib_basic():
    im = Image.init_blank([11, 11], fill=0, n_channels=1)
    im.landmarks['test'] = centre
    new_im = rasterize_landmarks_2d(im, group='test', render_lines=False,
                                    marker_style='.', marker_face_colour='r',
                                    marker_size=2, marker_edge_width=0,
                                    backend='matplotlib')
    assert new_im.n_channels == 3
    assert new_im.shape == (11, 11)
    assert_allclose(new_im.pixels[:, 5, 5], [255, 0, 0])


@unittest.skipIf(os.environ.get('TRAVIS'), 'Skipping due to lack of DISPLAY on Travis')
def test_rasterize_matplotlib_basic_line():
    im = Image.init_blank([11, 11], fill=0, n_channels=1)
    im.landmarks['test'] = line
    new_im = rasterize_landmarks_2d(im, group='test', render_lines=True,
                                    marker_style='.', marker_face_colour='r',
                                    marker_size=2, marker_edge_width=0,
                                    backend='matplotlib')
    assert new_im.n_channels == 3
    assert new_im.shape == (11, 11)
    assert_allclose(new_im.pixels[0, 3, 5], [255])
    assert_allclose(new_im.pixels[0, 9, 5], [255])
    assert_allclose(new_im.pixels[2, 5:8, 5], [255, 255, 255])


@unittest.skipIf(os.environ.get('TRAVIS'), 'Skipping due to lack of DISPLAY on Travis')
def test_rasterize_pillow_basic():
    im = Image.init_blank([11, 11], fill=0, n_channels=3)
    im.landmarks['test'] = centre
    new_im = rasterize_landmarks_2d(im, group='test', render_lines=False,
                                    marker_style='s', marker_face_colour='r',
                                    marker_size=1, marker_edge_width=0,
                                    backend='pillow')
    assert new_im.n_channels == 3
    assert new_im.shape == (11, 11)
    assert_allclose(new_im.pixels[0, 3:6, 3:6], 255)


@unittest.skipIf(os.environ.get('TRAVIS'), 'Skipping due to lack of DISPLAY on Travis')
def test_rasterize_pillow_basic_line():
    im = Image.init_blank([11, 11], fill=0, n_channels=3)
    im.landmarks['test'] = line
    new_im = rasterize_landmarks_2d(im, group='test', render_lines=True,
                                    line_width=1, line_colour='b',
                                    marker_style='s', marker_face_colour='r',
                                    marker_size=1, marker_edge_width=0,
                                    backend='pillow')
    assert new_im.n_channels == 3
    assert_allclose(new_im.pixels[0, 1:4, 3:6], 255)
    assert_allclose(new_im.pixels[0, 7:-1, 3:6], 255)
    assert_allclose(new_im.pixels[2, 4:7, 4], 255)



from collections import OrderedDict
from functools import wraps
import numpy as np

from menpo.base import name_of_callable
from menpo.landmark.base import LandmarkGroup
from menpo.landmark.exceptions import LabellingError


def connectivity_from_array(array, close_loop=False):
    r"""
    Build the connectivity over a given array. For example, given ::

        array = [(0, 3, 1, 2 )]

    Generate the connectivity of ::

        [(0, 3), (3, 1), (1, 2)]

    If ``close_loop`` is true, add an extra connection from the last point to
    the first.
    """
    # zip is a generator - need a list in this case
    conn = list(zip(array, array[1:]))
    if close_loop:
        conn.append((array[-1], array[0]))
    return np.asarray(conn)


def connectivity_from_range(range_tuple, close_loop=False):
    r"""
    Build the connectivity over a range. For example, given ::

        range_array = np.arange(3)

    Generate the connectivity of ::

        [(0, 1), (1, 2), (2, 3)]

    If ``close_loop`` is true, add an extra connection from the last point to
    the first.
    """
    return connectivity_from_array(
        np.arange(*range_tuple), close_loop=close_loop)


def validate_input(pcloud, n_expected_points):
    r"""
    Ensure that the input matches the number of expected points.

    Parameters
    ----------
    pcloud : :map:`PointCloud`
        Input Pointcloud to validate
    n_expected_points : `int`
        Number of expected points

    Raises
    ------
    LabellingError
        If the number of expected points doesn't match the number of given
        points
    """
    n_actual_points = pcloud.n_points
    if n_actual_points != n_expected_points:
        msg = 'Label expects exactly {} ' \
              'points. However, the given landmark group ' \
              'has {} points'.format(n_expected_points, n_actual_points)
        raise LabellingError(msg)


def pcloud_and_lgroup_from_ranges(pointcloud, labels_to_ranges):
    """
    Label the given pointcloud according to the given ordered dictionary
    of labels to ranges. This assumes that you can semantically label the group
    by using ranges in to the existing points e.g ::

        labels_to_ranges = {'jaw': (0, 17, False)}

    The third element of the range tuple is whether the range is a closed loop
    or not. For example, for an eye landmark this would be ``True``, as you
    do want to create a closed loop for an eye.

    Parameters
    ----------
    pointcloud : :map:`PointCloud`
        The pointcloud to apply semantic labels to.
    labels_to_ranges : `ordereddict` {`str` -> (`int`, `int`, `bool`)}
        Ordered dictionary of string labels to range tuples.

    Returns
    -------
    new_pcloud : :map:`PointCloud`
        New pointcloud with specific connectivity information applied.
    mapping : `ordereddict` {`str` -> `int ndarray`}
        For each label, the indices in to the pointcloud that belong to the
        label.
    """
    from menpo.shape import PointUndirectedGraph

    mapping = OrderedDict()
    all_connectivity = []
    for label, tup in labels_to_ranges.items():
        range_tuple = tup[:-1]
        close_loop = tup[-1]

        connectivity = connectivity_from_range(range_tuple,
                                               close_loop=close_loop)
        all_connectivity.append(connectivity)
        mapping[label] = np.arange(*range_tuple)
    all_connectivity = np.vstack(all_connectivity)

    new_pcloud = PointUndirectedGraph.init_from_edges(pointcloud.points,
                                                      all_connectivity)

    return new_pcloud, mapping


_labeller_docs = r"""
    Parameters
    ----------
    x : :map:`LandmarkGroup` or :map:`PointCloud` or `ndarray`
        The input landmark group, pointcloud or array to label. If a pointcloud
        is passed, then only the connectivity information is propagated to
        the pointcloud (a subclass of :map:`PointCloud` may be returned).
    return_mapping : `bool`, optional
        Only applicable if a :map:`PointCloud` or `ndarray` is passed. Returns
        the mapping dictionary which maps labels to indices into the resulting
        :map:`PointCloud` (which is then used to for building a
        :map:`LandmarkGroup`. This parameter is only provided for internal
        use so that other labellers can piggyback off one another.

    Returns
    -------
    x_labelled : :map:`LandmarkGroup` or :map:`PointCloud`
        If a :map:`LandmarkGroup` was passed, a :map:`LandmarkGroup` is
        returned. This landmark group will contain specific labels and
        these labels may refer to sub-pointclouds with specific connectivity
        information.

        If a :map:`PointCloud` was passed, a :map:`PointCloud` is returned. Only
        the connectivity information is propagated to the pointcloud
        (a subclass of :map:`PointCloud` may be returned).
    mapping_dict : `ordereddict` {`str` -> `int ndarray`}, optional
        Only returned if ``return_mapping==True``. Used for building
        :map:`LandmarkGroup`.

    Raises
    ------
    : :map:`LabellingError`
        If the given landmark group/pointcloud contains less than the
        expected number of points.
"""


def labeller_func(group_label=None):
    r"""
    Decorator for labelling functions. Labelling functions should return
    a template pointcloud and a mapping dictionary (from labels to indices).
    This decorator then takes that output and returns the correctly labelled
    object depending on if a landmark group, pointcloud or numpy array is
    passed. See the docstrings of the labelling methods where this
    has been made clear.

    Note that we duck type the group label (usually just the name of the
    labelling function) on to the function itself for the labeller method
    below.
    """
    def decorator(labelling_method):
        # Shadowing parent scope variables inside a nested function
        # kills the scope of the parent variable, so we need a unique alias
        # for the group name
        gl = (group_label if group_label is not None
              else name_of_callable(labelling_method))
        # Duck type group label onto method itself
        labelling_method.group_label = gl
        # Set up the global docs
        labelling_method.__doc__ += _labeller_docs

        @wraps(labelling_method)
        def wrapper(x, return_mapping=False):
            from menpo.shape import PointCloud
            # Accepts LandmarkGroup, PointCloud or ndarray
            if isinstance(x, np.ndarray):
                x = PointCloud(x, copy=False)

            if isinstance(x, PointCloud):
                new_pcloud, mapping = labelling_method(x)
                # This parameter is only provided for internal use so that
                # other labellers can piggyback off one another
                if return_mapping:
                    return new_pcloud, mapping
                else:
                    return new_pcloud
            if isinstance(x, LandmarkGroup):
                new_pcloud, mapping = labelling_method(x.lms)
                return LandmarkGroup.init_from_indices_mapping(new_pcloud, 
                                                               mapping)
        return wrapper
    return decorator


def labeller(landmarkable, group, label_func):
    """
    Re-label an existing landmark group on a :map:`Landmarkable` object with a
    new label set.

    Parameters
    ----------
    landmarkable: :map:`Landmarkable`
        :map:`Landmarkable` that will have it's :map:`LandmarkManager`
        augmented with a new :map:`LandmarkGroup`
    group: `str`
        The group label of the existing landmark group that should be
        re-labelled. A copy of this group will be attached to it's landmark
        manager with new labels. The group label of this new group and the
        labels it will have is determined by ``label_func``
    label_func: `func`  -> `(str, LandmarkGroup)`
        A labelling function taken from this module, Takes as input a
        :map:`LandmarkGroup` and returns a tuple of
        (new group label, new LandmarkGroup with semantic labels applied).

    Returns
    -------
    landmarkable : :map:`Landmarkable`
        Augmented ``landmarkable`` (this is just for convenience,
        the object will actually be modified in place)
    """
    new_group = label_func(landmarkable.landmarks[group])
    landmarkable.landmarks[label_func.group_label] = new_group
    return landmarkable



import warnings

import numpy as np
from numpy.testing import assert_allclose
from nose.tools import raises
from pathlib import Path

import menpo
from menpo.image import Image, MaskedImage, BooleanImage
from menpo.shape import PointCloud
from menpo.transform import UniformScale, Translation


def test_image_as_masked():
    img = Image(np.random.rand(3, 3, 1), copy=False)
    m_img = img.as_masked()
    assert(type(m_img) == MaskedImage)
    assert_allclose(m_img.pixels, img.pixels)


def test_image_has_nan_values():
    img = Image(np.random.rand(1, 3, 3), copy=False)
    img.pixels[0, 0, 0] = np.nan
    assert img.has_nan_values()


def test_image_no_nan_values():
    img = Image(np.random.rand(1, 3, 3), copy=False)
    assert not img.has_nan_values()


def test_masked_image_as_unmasked():
    m_img = MaskedImage(np.random.rand(1, 3, 3), copy=False)
    img = m_img.as_unmasked()
    assert(type(img) == Image)
    assert_allclose(m_img.pixels, img.pixels)


def test_masked_image_as_unmasked_fill():
    m_img = MaskedImage(np.random.rand(1, 3, 3), copy=False)
    m_img.mask.pixels[0, 0, 0] = False
    img = m_img.as_unmasked(fill=8)
    assert(type(img) == Image)
    assert_allclose(m_img.pixels[0, 1:, 1:], img.pixels[0, 1:, 1:])
    assert_allclose(img.pixels[0, 0, 0], 8.0)


def test_masked_image_as_unmasked_fill_tuple():
    m_img = MaskedImage(np.random.rand(3, 3, 3), copy=False)
    m_img.mask.pixels[0, 0, 0] = False
    img = m_img.as_unmasked(fill=(1, 2, 3))
    assert(type(img) == Image)
    assert_allclose(m_img.pixels[0, 1:, 1:], img.pixels[0, 1:, 1:])
    assert_allclose(img.pixels[:, 0, 0], (1, 2, 3))


@raises(NotImplementedError)
def test_boolean_image_as_masked_raises_not_implemented_error():
    b_img = BooleanImage.init_blank((4, 5))
    b_img.as_masked()


def test_warp_to_shape_preserves_path():
    bb = menpo.io.import_builtin_asset.breakingbad_jpg()
    bb2 = bb.rescale(0.1)
    assert hasattr(bb2, 'path')
    assert bb2.path == bb.path


def test_warp_to_mask_preserves_path():
    bb = menpo.io.import_builtin_asset.breakingbad_jpg()
    no_op = UniformScale(1.0, n_dims=2)
    bb2 = bb.warp_to_mask(BooleanImage.init_blank((10, 10)), no_op)
    assert hasattr(bb2, 'path')
    assert bb2.path == bb.path


def test_warp_to_shape_boolean_preserves_path():
    i1 = BooleanImage.init_blank((10, 10))
    i1.path = Path('.')
    i2 = i1.rescale(0.8)
    assert hasattr(i2, 'path')
    assert i2.path == i1.path


def test_init_from_rolled_channels():
    p = np.empty([50, 60, 3])
    im = Image.init_from_channels_at_back(p)
    assert im.n_channels == 3
    assert im.height == 50
    assert im.width == 60


def test_init_from_pointcloud():
    pc = PointCloud.init_2d_grid((10, 10))
    im = Image.init_from_pointcloud(pc)
    assert im.shape == (9, 9)


def test_init_from_pointcloud_return_transform():
    correct_tr = Translation([5, 5])
    pc = correct_tr.apply(PointCloud.init_2d_grid((10, 10)))
    im, tr = Image.init_from_pointcloud(pc, return_transform=True)
    assert im.shape == (9, 9)
    assert_allclose(tr.as_vector(), -correct_tr.as_vector())


def test_init_from_pointcloud_attach_group():
    pc = PointCloud.init_2d_grid((10, 10))
    im = Image.init_from_pointcloud(pc, group='test')
    assert im.shape == (9, 9)
    assert im.n_landmark_groups == 1


def test_init_from_pointcloud_boundary():
    pc = PointCloud.init_2d_grid((10, 10))
    im = Image.init_from_pointcloud(pc, boundary=5)
    print(im.shape)
    assert im.shape == (19, 19)


def test_bounds_2d():
    im = Image.init_blank((50, 30))
    assert_allclose(im.bounds(), ((0, 0), (49, 29)))


def test_bounds_3d():
    im = Image.init_blank((50, 30, 10))
    assert_allclose(im.bounds(), ((0, 0, 0), (49, 29, 9)))


def test_constrain_landmarks_to_bounds():
    im = Image.init_blank((10, 10))
    im.landmarks['test'] = PointCloud.init_2d_grid((20, 20))
    with warnings.catch_warnings():
        warnings.simplefilter('ignore')
        im.constrain_landmarks_to_bounds()
    assert not im.has_landmarks_outside_bounds()
    assert_allclose(im.landmarks['test'].lms.bounds(), im.bounds())



import collections

from mock import Mock
from nose.tools import raises

from menpo.base import LazyList


def test_lazylist_get():
    mock_func = Mock()
    mock_func.return_value = 1
    ll = LazyList([mock_func] * 10)
    assert len(ll) == 10
    assert ll[0] == 1


def test_lazylist_copy_lazy():
    mock_func = Mock()
    mock_func.return_value = 1
    ll = LazyList([mock_func] * 10)
    copied_ll = ll.copy()
    assert len(copied_ll) == 10
    assert id(ll._callables) != id(copied_ll._callables)
    mock_func.assert_not_called()


def test_lazylist_copy_duck_typed():
    mock_func = Mock()
    mock_func.return_value = 1
    ll = LazyList([mock_func] * 10)
    ll.fps = 50
    copied_ll = ll.copy()
    assert len(copied_ll) == 10
    assert id(ll._callables) != id(copied_ll._callables)
    mock_func.assert_not_called()
    assert copied_ll.fps == 50


def test_lazylist_multiple_calls():
    mock_func = Mock()
    mock_func.return_value = 1
    ll = LazyList([mock_func] * 10)
    ll[0]
    ll[0]
    assert mock_func.call_count == 2


def test_lazylist_multi_map():
    two_func = lambda: 2
    double_func = [lambda x: x * 2] * 2
    ll = LazyList([two_func] * 2)
    ll_mapped = ll.map(double_func)
    assert len(ll_mapped) == 2
    assert id(ll) != id(ll_mapped)
    assert all(x == 4 for x in ll_mapped)


@raises(ValueError)
def test_lazylist_multi_map_unequal_lengths():
    two_func = lambda: 2
    double_func = [lambda x: x * 2] * 2
    ll = LazyList([two_func])
    ll.map(double_func)


@raises(ValueError)
def test_lazylist_multi_map_iterable_and_callable():
    class double_func(collections.Iterable):
        def __call__(self, x, **kwargs):
            return x * 2

        def __iter__(self):
            yield 1

    f = double_func()
    two_func = lambda: 2
    ll = LazyList([two_func])
    ll.map(f)
    assert 1


def test_lazylist_repeat():
    ll = LazyList.init_from_iterable([0, 1])
    ll_repeated = ll.repeat(2)
    assert len(ll_repeated) == 4
    assert all([a == b for a, b in zip([0, 0, 1, 1], ll_repeated)])


def test_lazylist_map():
    two_func = lambda: 2
    double_func = lambda x: x * 2
    ll = LazyList([two_func])
    ll_mapped = ll.map(double_func)
    assert id(ll) != id(ll_mapped)
    assert ll_mapped[0] == 4


def test_lazylist_double_map():
    two_func = lambda: 2
    double_func = lambda x: x * 2
    ll = LazyList([two_func])
    ll_mapped = ll.map(double_func)
    ll_mapped = ll_mapped.map(double_func)
    assert id(ll) != id(ll_mapped)
    assert ll_mapped[0] == 8


def test_lazylist_map_no_call():
    mock_func = Mock()
    double_func = lambda x: x * 2
    ll = LazyList([mock_func])
    ll_mapped = ll.map(double_func)
    assert id(ll) != id(ll_mapped)
    mock_func.assert_not_called()


def test_lazylist_init_from_iterable_identity():
    ll = LazyList.init_from_iterable([0, 1])
    assert ll[0] == 0
    assert ll[1] == 1


def test_lazylist_init_from_iterable_with_f():
    double_func = lambda x: x * 2
    ll = LazyList.init_from_iterable([0, 1], f=double_func)
    assert ll[0] == 0
    assert ll[1] == 2


def test_lazylist_init_from_index_callable():
    identity_func = lambda x: x
    ll = LazyList.init_from_index_callable(identity_func, 5)
    assert ll[0] == 0
    assert ll[-1] == 4


@raises(TypeError)
def test_lazylist_immutable():
    ll = LazyList([])
    ll[0] = 1


def test_lazylist_add_lazylist():
    a = Mock()
    b = Mock()
    ll1 = LazyList([a])
    ll2 = LazyList([b])
    new_ll = ll1 + ll2
    assert len(new_ll) == 2
    assert new_ll._callables[0] is a
    assert new_ll._callables[1] is b


def test_lazylist_add_list():
    a = Mock()
    b = Mock()
    ll1 = LazyList([a])
    l2 = [b]
    new_ll = ll1 + l2
    assert len(new_ll) == 2
    assert new_ll._callables[0] is a
    assert new_ll._callables[1] is not b
    assert new_ll[1] is b



# distutils: language = c++

import numpy as np
cimport numpy as np
cimport cython


ctypedef fused DOUBLE_TYPES:
    float
    double


cdef extern from "cpp/central_difference.h":
    void central_difference[T](const T* input, const Py_ssize_t rows,
                               const Py_ssize_t cols, const Py_ssize_t n_channels,
                               T* output)


@cython.boundscheck(False)
@cython.wraparound(False)
cpdef gradient_cython(np.ndarray[DOUBLE_TYPES, ndim=3] input):

    cdef Py_ssize_t n_channels = input.shape[0]
    cdef Py_ssize_t rows = input.shape[1]
    cdef Py_ssize_t cols = input.shape[2]
    # Maintain the dtype that was passed in (float or double)
    dtype = input.dtype
    cdef np.ndarray[DOUBLE_TYPES, ndim=3] output = np.zeros((n_channels * 2,
                                                            rows, cols),
                                                            dtype=dtype)

    central_difference(&input[0,0,0], rows, cols, n_channels,
                       &output[0,0,0])

    return output



import numpy as np
import cython
cimport numpy as np
cimport cython

ctypedef np.float64_t FLOAT64_T
ctypedef fused integrals:
    np.uint32_t
    np.uint64_t
    np.int32_t
    np.int64_t

cdef np.ndarray[FLOAT64_T, ndim=2] normalize(np.ndarray[FLOAT64_T, ndim=2] vec):
    """
    Normalize the given array of vectors.

    Parameters
    ----------
    vec : (N, 3) c-contiguous double ndarray
        The array of vectors to normalize

    Returns
    -------
    normalized : (N, 3) c-contiguous double ndarray
        Normalized array of vectors.
    """
    # Avoid divisions by almost 0 numbers
    # np.spacing(1) is equivalent to Matlab's eps
    cdef np.ndarray[FLOAT64_T, ndim=1] d = np.sqrt(np.sum(vec ** 2, axis=1))
    d[d < np.spacing(1)] = 1.0
    return vec / d[..., None]


cdef inline np.ndarray[FLOAT64_T, ndim=2] cross(double[:, :] x,
                                                double[:, :] y):
    """
    The N x 3 cross product (returns the vectors orthogonal
    to ``x`` and ``y``). This performs the cross product on each (3, 1) vector
    in the two arrays. Assumes ``x`` and ``y`` have the same shape.

    Parameters
    ----------
    x : (N, 3) double memory view
        First array to perform cross product with.
    y : (N, 3) double memory view
        Second array to perform cross product with.

    Returns
    -------
    cross : (N, 3) c-contiguous double ndarray
        The array of vectors representing the cross product between each
        corresponding vector.
    """
    cdef:
        np.ndarray[FLOAT64_T, ndim=2] z = np.empty_like(x)
        Py_ssize_t n = x.shape[0], i = 0
    for i in range(n):
        z[i, 0] = x[i, 1] * y[i, 2] - x[i, 2] * y[i, 1]
        z[i, 1] = x[i, 2] * y[i, 0] - x[i, 0] * y[i, 2]
        z[i, 2] = x[i, 0] * y[i, 1] - x[i, 1] * y[i, 0]

    return z


cpdef compute_normals(np.ndarray[FLOAT64_T, ndim=2] vertex,
                      np.ndarray[integrals, ndim=2] face):
    """
    Compute the per-vertex and per-face normal of the vertices given a list of
    faces. Ensures that all the normals are pointing in a consistent direction
    (to avoid 'inverted' normals).

    Parameters
    ----------
    vertex : (N, 3) c-contiguous double ndarray
        The list of points to compute normals for.
    face : (M, 3) c-contiguous double ndarray
        The list of faces (triangle list).

    Returns
    -------
    vertex_normal : (N, 3) c-contiguous double ndarray
        The normal per vertex.
    face_normal : (M, 3) c-contiguous double ndarray
        The normal per face.
    :return:
    """
    cdef int nface = face.shape[0]
    cdef int nvert = vertex.shape[0]

    # Calculate the cross product (per-face normal)
    cdef np.ndarray[FLOAT64_T, ndim=2] face_normal = cross(
        vertex[face[:, 1], :] - vertex[face[:, 0], :],
        vertex[face[:, 2], :] - vertex[face[:, 0], :])
    face_normal = normalize(face_normal)

    # Calculate per-vertex normal
    cdef np.ndarray[FLOAT64_T, ndim=2] vertex_normal = np.zeros([nvert, 3])
    cdef integrals f0, f1, f2
    for i in range(nface):
        f0 = face[i, 0]
        f1 = face[i, 1]
        f2 = face[i, 2]
        for j in range(3):
            vertex_normal[f0, j] += face_normal[i, j]
            vertex_normal[f1, j] += face_normal[i, j]
            vertex_normal[f2, j] += face_normal[i, j]

    # Normalize
    vertex_normal = normalize(vertex_normal)

    return vertex_normal, face_normal



from collections import OrderedDict
import numpy as np

from ..base import validate_input, connectivity_from_array, labeller_func


@labeller_func(group_label='hand_ibug_39')
def hand_ibug_39_to_hand_ibug_39(pcloud):
    r"""
    Apply the IBUG 39-point semantic labels.

    The semantic labels applied are as follows:

      - thumb
      - index
      - middle
      - ring
      - pinky
      - palm
    """
    from menpo.shape import PointUndirectedGraph

    n_expected_points = 39
    validate_input(pcloud, n_expected_points)

    thumb_indices = np.arange(0, 5)
    index_indices = np.arange(5, 12)
    middle_indices = np.arange(12, 19)
    ring_indices = np.arange(19, 26)
    pinky_indices = np.arange(26, 33)
    palm_indices = np.hstack((np.array([32, 25, 18, 11, 33, 34, 4]),
                              np.arange(35, 39)))

    thumb_connectivity = connectivity_from_array(thumb_indices,
                                                 close_loop=False)
    index_connectivity = connectivity_from_array(index_indices,
                                                 close_loop=False)
    middle_connectivity = connectivity_from_array(middle_indices,
                                                  close_loop=False)
    ring_connectivity = connectivity_from_array(ring_indices,
                                                close_loop=False)
    pinky_connectivity = connectivity_from_array(pinky_indices,
                                                 close_loop=False)
    palm_connectivity = connectivity_from_array(palm_indices,
                                                close_loop=True)

    all_connectivity = np.vstack([thumb_connectivity, index_connectivity,
                                  middle_connectivity, ring_connectivity,
                                  pinky_connectivity, palm_connectivity])

    new_pcloud = PointUndirectedGraph.init_from_edges(pcloud.points,
                                                      all_connectivity)

    mapping = OrderedDict()
    mapping['thumb'] = thumb_indices
    mapping['index'] = index_indices
    mapping['middle'] = middle_indices
    mapping['ring'] = ring_indices
    mapping['pinky'] = pinky_indices
    mapping['palm'] = palm_indices

    return new_pcloud, mapping
